from __future__ import unicode_literals
from builtins import input
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from automate.common import ExitException
from automate.service import AbstractUserService

helpstr = """
 Automate interactive Python shell
 ==================================

 This interactive shell is intended for making small changes to
 working model.

 Any Python commands are accepted. You may refer to all your
 automate objects by their name attribute. You may do anything to
 modify the state on fly by this field, for example, create new programs,
 actuators, sensors, etc. Some practical commands:

  help -- show this help message (or with parameter, normal pydoc)
  lsa, lsp, lss -- list actuators, programs, sensors
  get_statusmsg -- print current system status
  gui -- show GUI
  quit -- quit cleanly

  python_str(filename) -- print all objects in a python file
  obj.set_value(value) -- set object value for obj
  obj.print_traits() -- print all traits attributes for obj

"""

prompt = "> "


class TextUIService(AbstractUserService):

    """
        Provides interactive Python shell frontend to the System.
        Uses IPython if it is installed.
        Provides couple of functions to the System namespace.
    """
    autoload = True

    def setup(self):
        self.system.register_service_functions(
            self.ls, self.lsa, self.lsp, self.lss, self.help, self.text_ui, self.quit)
        self.system.on_trait_change(self.text_ui, 'post_init_trigger')

    def ls(self, what):
        """List actuators, programs or sensors (what is string)"""
        for i in getattr(self.system, what):
            self.logger.info('%s: %s: %s', i.__class__.__name__, i, i.status)
        return True

    def lsa(self):
        """List actuators"""
        return self.ls("actuators")

    def lsp(self):
        """List programs"""
        return self.ls("programs")

    def lss(self):
        """List sensors"""
        return self.ls("sensors")

    def help(self, *args, **kwargs):
        """Print Automate help if no parameter is given. Otherwise,
           act as pydoc.help()"""
        if len(args) > 0 or len(kwargs) > 0:
            import pydoc

            pydoc.help(*args, **kwargs)
        else:
            hstr = helpstr
            for i in hstr.split("\n"):
                self.logger.info(i)
        return True

    def text_ui(self):
        """
            Start Text UI main loop
        """
        self.logger.info("Starting command line interface")
        self.help()
        try:
            self.ipython_ui()
        except ImportError:
            self.fallback_ui()
        self.system.cleanup()

    def ipython_ui(self):
        from IPython.terminal.ipapp import TerminalIPythonApp
        import automate
        self.system.namespace.allow_overwrite.extend(['_', '__', '___', '_i', '_ii', '_iii', 'quit'])
        self.system.namespace.update({k: v for k, v in list(automate.__dict__.items()) if k not in self.system.namespace})
        term = TerminalIPythonApp(user_ns=self.system.namespace)
        self.system.namespace['term'] = term
        term.initialize([])
        term.start()

    def fallback_ui(self):
        # If IPython is not installed, this fallback is used
        try:
            while True:
                try:
                    import readline
                    import rlcompleter

                    rlcompleter.Completer._callable_postfix = lambda self, val, word: word
                    readline.set_completer(rlcompleter.Completer(self.system.cmd_namespace).complete)
                    readline.parse_and_bind("tab: complete")
                except ImportError:
                    self.logger.warning(
                        "Readline support disabled. Please install readline and rlcompleter if you want to use.")
                try:
                    c = eval(input(prompt))
                except EOFError:
                    return
                if c[:1] == "%":
                    c = c[1:]
                else:
                    cmd = c.strip()
                    self.system.cmd_exec(cmd)

        except (KeyboardInterrupt, ExitException):
            return

    def quit(self):
        raise ExitException



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-arduino.
#
# automate-arduino is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-arduino is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-arduino.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from .arduino_service import ArduinoService
from .arduino_actuators import ArduinoDigitalActuator, ArduinoPWMActuator, ArduinoServoActuator, AbstractArduinoActuator
from .arduino_sensors import ArduinoAnalogSensor, ArduinoDigitalSensor, AbstractArduinoSensor

extension_classes = [ArduinoService, ArduinoDigitalActuator, ArduinoPWMActuator, ArduinoServoActuator, AbstractArduinoActuator,
                     ArduinoAnalogSensor, ArduinoDigitalSensor, AbstractArduinoSensor, ]



from __future__ import unicode_literals
from builtins import object
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

import pytest

from automate import *
from automate.program import Program


@pytest.fixture(params=[Program, UserBoolSensor, BoolActuator])
def progtype(request):
    return request.param


@pytest.fixture(params=[0, 1])
def mysys(request, progtype):
    class sys(System):
        s1 = UserBoolSensor()
        s2 = UserBoolSensor()
        a1 = BoolActuator(default=False)
        a2 = BoolActuator()
        p = progtype()
        p1 = progtype(priority=1.0, active_condition=Value(True))
        p2 = progtype(priority=2.0, active_condition=Value(True))
        if request.param == 0:
            p3 = progtype(priority=2.0, active_condition=Value(False),
                          on_activate=Run(Run(SetStatus(a2, s1),)))
        else:
            p3 = progtype(priority=2.0, active_condition=Value(False),
                          on_activate=Run(Run(SetStatus('a2', 's1'))),
                          )
    s = sys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()


def test_correct_callable_setup(sysloader):
    class ms(System):

        class grp2(Group):
            s = UserBoolSensor()
            s2 = UserBoolSensor()
            a = BoolActuator()

        b = BoolActuator()

        class grp(Group):
            p = Program(priority=2.0,
                        active_condition=And(Or('s', 's2')),
                        on_activate=Run(Run(SetStatus('a', 's'))),
                        update_condition=Run(Run(SetStatus('a', 's'))),
                        )

    s = sysloader.new_system(ms)
    p = s.p
    assert isinstance(p.on_activate[0][0][0], BoolActuator)
    assert isinstance(p.on_activate[0][0][1], UserBoolSensor)
    assert isinstance(p.update_condition[0][0][0], BoolActuator)
    assert isinstance(p.update_condition[0][0][1], UserBoolSensor)
    assert isinstance(p.active_condition[0][0], UserBoolSensor)
    assert isinstance(p.active_condition[0][1], UserBoolSensor)


def test_update_on_activate(mysys):
    a1 = mysys.a1
    dp = a1.default_program
    assert a1.program_status[dp] == False
    assert repr(dp.on_activate) == "SetStatus('a1', Value(False))"
    dp.on_activate = SetStatus(a1, True)
    mysys.flush()
    assert a1.program_status[dp] == True


@pytest.fixture
def freezesys(request):
    class sys(System):
        s1 = UserBoolSensor()
        p1 = Program(
            active_condition=Value(s1),
            on_activate=SetStatus(s1, 1)
        )
    s = sys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()


@pytest.fixture(params=[UserBoolSensor, BoolActuator])
def freezesys2(request):
    class sys(System):
        s = request.param()
        p = Program(on_update=SetStatus(s, 1), triggers=[s])

    s = sys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()


def test_freeze2(freezesys2):
    assert freezesys2


def test_freeze(freezesys):
    freezesys.s1.status = 1
    freezesys.flush()
    assert freezesys.s1.status == True


@pytest.fixture
def freezesys_act(request):
    class sys(System):
        s1 = UserBoolSensor()
        a1 = BoolActuator()
        a2 = BoolActuator()
        p1 = Program(
            update_condition=Or(s1, a1),
            on_activate=Run(
                SetStatus(a1, s1),
                SetStatus(a2, a1),
                SetStatus(a1, s1),
                SetStatus(a2, a1),
            )
        )
    s = sys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()

#@pytest.mark.skipif(True)


def test_freeze_act(freezesys_act):
    freezesys = freezesys_act
    freezesys.s1.status = 1
    freezesys.flush()
    assert freezesys.s1.status == True


class TestProgramFeatures(object):

    def test_triggerlist_targetlist_change(self, mysys):
        p = mysys.p
        #add = {p} if isinstance(p, StatusObject) else set()
        add = set()

        assert p.actual_triggers == set() | add
        p.active_condition = Value(mysys.s1)
        assert p.actual_triggers == {mysys.s1} | add
        p.on_activate = SetStatus(mysys.a1, Value(True))
        assert p.actual_triggers == {mysys.s1} | add
        assert p.actual_targets == {mysys.a1}
        assert mysys.a1.status == False
        mysys.s1.set_status(True)
        mysys.worker_thread.flush()
        assert mysys.a1.status == True
        p.targets = [mysys.a2]
        assert set(p.actual_targets) == {mysys.a1, mysys.a2}
        p.triggers = [mysys.s2]
        assert set(p.actual_triggers) == {mysys.s1, mysys.s2}
        p.triggers = []
        p.targets = []
        assert set(p.actual_targets) == {mysys.a1}
        assert set(p.actual_triggers) == {mysys.s1}
        p.targets = ['a2']
        assert set(p.actual_targets) == {mysys.a1, mysys.a2}
        p.triggers = ['s2']
        assert set(p.actual_triggers) == {mysys.s1, mysys.s2}
        p.exclude_triggers = ['s1']
        assert set(p.actual_triggers) == {mysys.s2}
        p.exclude_triggers = [mysys.s1]
        assert set(p.actual_triggers) == {mysys.s2}

    def test_triggerlist_targetlist_change2(self, mysys):
        p = mysys.p
        #add = {p} if isinstance(p, StatusObject) else set()
        add = set()
        assert set(p.actual_triggers) == add
        for i in ['on_update', 'on_activate', 'update_condition', 'on_deactivate', 'active_condition']:
            setattr(p, i, Value(mysys.s1))
            assert p.actual_triggers == {mysys.s1} | add
        for i in ['on_update', 'on_activate', 'on_deactivate']:
            setattr(p, i, SetStatus(mysys.a1, mysys.s1))
            assert p.actual_triggers == {mysys.s1} | add
            assert p.actual_targets == {mysys.a1}

    def test_triggerlist_targetlist_change2_namever(self, mysys):
        # TODO: investigate failure in https://travis-ci.org/tuomas2/automate/jobs/245124552
        # and                          https://travis-ci.org/tuomas2/automate/jobs/246254565
        #                              https://travis-ci.org/tuomas2/automate/jobs/247160904
        #                            N https://travis-ci.org/tuomas2/automate/jobs/251122776
        #                            N https://travis-ci.org/tuomas2/automate/jobs/251179301
        #                            N https://travis-ci.org/tuomas2/automate/jobs/251888696
        # TODO: check also similar tests above.
        p = mysys.p
        #add = {p} if isinstance(p, StatusObject) else set()
        add = set()
        assert set(p.actual_triggers) == add
        for i in ['on_update', 'on_activate', 'update_condition', 'on_deactivate', 'active_condition']:
            setattr(p, i, Value('s1'))
            assert p.actual_triggers == {mysys.s1} | add
        for i in ['on_update', 'on_activate', 'on_deactivate']:
            setattr(p, i, SetStatus('a1', 's1'))
            assert p.actual_triggers == {mysys.s1} | add
            assert p.actual_targets == {mysys.a1}

    def test_program_priorities(self, mysys):
        p1, p2 = mysys.p1, mysys.p2
        dp = mysys.a1.program_stack[0]
        mysys.s1.set_status(True)
        mysys.s2.set_status(False)
        mysys.worker_thread.flush()
        assert p1.actual_targets == set()
        p1.active = False  # on activate triggers only when activation
        # really changes. So, only changing on_activate doesn't trigger
        # it unless deactivation takes place. It's correct
        # this way, I think.

        p1.on_activate = SetStatus(mysys.a1, mysys.s1)
        mysys.flush()
        assert p1.active
        assert p1.actual_targets == {mysys.a1}
        assert mysys.a1.program_stack == [dp, p1]
        assert mysys.a1.program == p1
        assert mysys.a1.status == mysys.s1.status == True

        p2.active = False
        p2.on_activate = SetStatus(mysys.a1, mysys.s2)
        assert mysys.a1.program == p2
        assert mysys.a1.program_stack == [dp, p1, p2]
        mysys.flush()
        assert mysys.a1.status == False
        p2.priority = 0.5
        assert mysys.a1.program == p1
        assert mysys.a1.program_stack == [dp, p2, p1]
        mysys.flush()
        assert mysys.a1.status == True
        p1.priority = -1
        p2.priority = -2
        assert mysys.a1.program == dp
        assert mysys.a1.program_stack == [p2, p1, dp]
        mysys.flush()
        assert mysys.a1.status == False


def test_logicstr(mysys):
    assert mysys.p2.active_condition_str == 'Value(True)'
    assert mysys.p2.active_condition == Value(True)
    mysys.p2.active_condition_str = 'Value(False)'
    assert mysys.p2.active_condition_str == 'Value(False)'
    assert mysys.p2.active_condition == Value(False)
    from traits.api import TraitError
    with pytest.raises(TraitError):
        mysys.p2.active_condition_str = 'asfdjk'
    assert mysys.p2.active_condition_str == 'Value(False)'

    assert mysys.p3.targets_str in ['TraitSetObject()', 'TraitSetObject([])']
    assert mysys.p3.targets == set([])
    mysys.p3.targets_str = '{a1}'
    assert mysys.p3.targets_str in ["TraitSetObject(['a1'])", "TraitSetObject({'a1'})"]
    assert mysys.p3.targets == {mysys.a1}



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
import platform
from threading import Thread

from traits.api import CUnicode, Unicode, CInt
import requests

from automate.common import threaded
from automate.systemobject import SystemObject


class PushOver(SystemObject):
    """
        Send push notifications to android/IOS phones (see http://pushover.net).

        Use this as you would use a Callable.
    """

    #: Pushover.net application/API token
    api_key = CUnicode
    #: Pushover.net user key
    user_key = CUnicode
    #: Pushover.net priority
    priority = CInt(0)
    #: Device name (empty to send to all devices)
    device = CUnicode
    #: Sound name in device
    sound = CUnicode

    def send(self, caller):
        try:
            url = self.system.services_by_name['WebService'][0].server_url
        except KeyError:
            url = ''

        data=dict(
            token=self.api_key,
            user=self.user_key,
            message='Automate event triggered on %s' % platform.node(),
            title=caller.name,
            url=url,
            url_title=self.system.name + ' web interface')

        if self.device:
            data['device'] = self.device
        if self.sound:
            data['sound'] = self.sound

        requests.post('https://api.pushover.net/1/messages.json', data=data)

    def call(self, caller, **kwargs):
        t = Thread(target=threaded(self.send, caller), name='Notification sender thread')
        t.start()


class EmailSender(SystemObject):

    """ Send email notification of the current status of the system """

    # Email address where mail is sent
    to_email = CUnicode
    smtp_hostname = Unicode
    smtp_username = Unicode
    smtp_password = Unicode
    smtp_fromname = Unicode
    smtp_fromemail = Unicode

    def get_status_display(self):
        return self.to_email

    def call(self, caller, **kwargs):
        t = Thread(target=threaded(self.send, caller), name='Email sender thread')
        t.start()

    def send(self, caller):
        self.logger.info("Sending email now")

        progname = self.system.name

        hostname = platform.node()
        subject = u'Program {progname} at {host} encountered event of "{evname}"! '.format(
            host=hostname, progname=progname, evname=caller.name)

        headers = (u"Subject: {subject}\n"
                   "From: {fromaddr}\n"
                   "To: {to}\n"
                   "Content-type: text/plain; charset=UTF-8\n").format(subject=subject,
                                                                       fromaddr=self.smtp_fromemail,
                                                                       to=self.to_email)

        sensorstatuses = '\n'.join('%s: %s' % (i.name, i.status) for i in self.system.sensors)
        actuatorstatuses = '\n'.join('%s: %s' % (i.name, i.status) for i in self.system.actuators)

        logserv = self.system.request_service('LogStoreService')
        message = (u'{headers}\n'
                   'Program {progname} encountered event of "{evname}"!\n\n'
                   'Sensors\n\n'
                   '{sensorstatuses}\n\n'
                   'Actuators\n\n'
                   '{actuatorstatuses}\n\n'
                   'Log\n\n{log}').format(progname=progname,
                                          evname=caller.name,
                                          headers=headers,
                                          sensorstatuses=sensorstatuses,
                                          actuatorstatuses=actuatorstatuses,
                                          log=logserv.lastlog(html=False))
        message = message

        import smtplib

        smtp = smtplib.SMTP_SSL(self.smtp_hostname)
        smtp.login(self.smtp_username, self.smtp_password)
        smtp.sendmail(self.smtp_fromemail, self.to_email, message)


class PlantUMLGenerator(SystemObject):

    def call(self, caller=None):
        s = self.system.request_service('PlantUMLService')
        return s.write_puml()



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

"""
    Module for various Sensor classes.
"""
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import filter

import time
import socket
import logging
import subprocess
import types
import pyinotify
import threading
import queue
from copy import deepcopy

from datetime import datetime, timedelta

from croniter import croniter
from traits.api import Any, CInt, CFloat, Unicode, CUnicode, List, CBool, Instance, CStr, Int

from automate.common import get_modules_all
from automate.common import threaded, Lock
from automate.statusobject import AbstractSensor
from automate.callables import Value
from automate.callable import AbstractCallable

logger = logging.getLogger('automate.sensor')


class UserAnySensor(AbstractSensor):

    """User editable sensor type that accepts values of any types"""
    user_editable = CBool(True)
    _status = Any


class UserBoolSensor(AbstractSensor):

    """Boolean-valued user-editable sensor"""
    user_editable = CBool(True)
    _status = CBool


class UserEventSensor(UserBoolSensor):

    """
        Boolean-valued user-editable sensor suitable for using for singular events.

        After status has been changed to ``True``, it changes automatically its status
        back to ``False``.
    """

    def __status_changed(self):
        self.status = False

    def get_default_callables(self):
        callables = super(UserEventSensor, self).get_default_callables()
        callables['active_condition'] = Value(self)
        return callables


class AbstractNumericSensor(AbstractSensor):

    """
        Abstract class for numeric sensor types, that allows limiting
        value within a specific range.

        If limiting values (:attr:`.value_min`, :attr:`.value_max`) are used, value that exceeds
        these limits, is clipped to the range.
    """

    #: Minimum allowed value for status
    value_min = CFloat(float('-inf'))

    #: Maximum allowed value for status
    value_max = CFloat(float('inf'))

    view = AbstractSensor.view + ['value_min', 'value_max']

    @property
    def is_finite_range(self):
        return self.value_max - self.value_min < float('inf')

    def get_as_datadict(self):
        d = super(AbstractNumericSensor, self).get_as_datadict()
        d.update(dict(value_min=self.value_min, value_max=self.value_max))
        return d

    def set_status(self, status, **kwargs):
        if status is None:
            clipped_status = None
        else:
            clipped_status = max(min(float(status), self.value_max), self.value_min)
        super(AbstractNumericSensor, self).set_status(clipped_status, **kwargs)


class UserIntSensor(AbstractNumericSensor):

    """Integer-valued user-editable sensor"""
    user_editable = CBool(True)
    _status = CInt(0)


class UserFloatSensor(AbstractNumericSensor):

    """Float-valued user-editable sensor"""
    user_editable = CBool(True)
    _status = CFloat
    silent = CBool(True)


class UserStrSensor(AbstractSensor):

    """String-valued user-editable sensor"""
    user_editable = CBool(True)
    _status = CUnicode


class CroniterOn(croniter):
    pass


class CroniterOff(croniter):
    pass


class CronTimerSensor(AbstractSensor):

    """
        Scheduled start/stop timer. Both start and stop times
        are configured by cron-type string (see man 5 crontab for description of the
        definition format).
    """
    class CronListStr(Unicode):

        """Validation class for cron-compatible strings (for timers)"""

        def validate(self, object, name, value):
            vals = value.split(";")
            for v in vals:
                try:
                    c = croniter(v)
                except:
                    self.error(object, name, value)
                    return
            return value

    _status = CBool(False)

    #: Semicolon separated lists of cron-compatible strings that indicate
    #: when to switch status to ``True``
    timer_on = CronListStr("0 0 0 0 0")

    #: Semicolon separated lists of cron-compatible strings that indicate
    #: when to switch status to ``False``
    timer_off = CronListStr("0 0 0 0 0")

    _update_timer = Any(transient=True)  # Timer object
    _timerlock = Any(transient=True)  # Lock object

    view = UserBoolSensor.view + ["timer_on", "timer_off"]

    def setup_system(self, *args, **traits):
        self._timerlock = Lock()
        super(CronTimerSensor, self).setup_system(*args, **traits)
        self.update_status()

    def _now(self):
        return datetime.now()

    def update_status(self):
        with self._timerlock:
            now = self._now()
            next_iters = [CroniterOn(i, now) for i in self.timer_on.split(";")] + \
                         [CroniterOff(i, now) for i in self.timer_off.split(";")]

            for i in next_iters:
                i.get_next(datetime)

            next_iters.sort(key=lambda x: x.get_current(datetime))

            prev_iters = deepcopy(next_iters)

            for i in prev_iters:
                i.get_prev(datetime)
            prev_iters.sort(key=lambda x: x.get_current(datetime))

            self.status = isinstance(prev_iters[-1], CroniterOn)

            self._setup_next_update(next_iters[0].get_current(datetime))

    def _timer_on_changed(self, name, new):
        self.update_status()

    def _timer_off_changed(self, name, new):
        self.update_status()

    def _setup_next_update(self, next_update_time):
        now = self._now()
        if self._update_timer and self._update_timer.is_alive():
            self._update_timer.cancel()

        delay = next_update_time - now + timedelta(seconds=5)
        self.logger.info('Setting timer to %s, %s seconds, at %s', delay, delay.seconds, now+delay)
        self._update_timer = threading.Timer(delay.seconds, threaded(self.update_status,))
        self._update_timer.name = ("Timer for TimerSensor %s at %s (%s seconds)" % (self.name, now + delay, delay.seconds))
        self._update_timer.start()

    def cleanup(self):
        with self._timerlock:
            if self._update_timer:
                self._update_timer.cancel()


class FileChangeSensor(AbstractSensor):

    """ Sensor that detects file changes on filesystem.
        Integer valued status is incremented by each change.
    """
    _status = CInt(0)

    #: Name of file or directory to monitor
    filename = CUnicode

    #: PyInotify flags to configure what file change events to monitor
    watch_flags = Int(pyinotify.IN_MODIFY | pyinotify.IN_CREATE | pyinotify.IN_DELETE)

    _notifier = Any(transient=True)

    class InotifyEventHandler(pyinotify.ProcessEvent):

        def __init__(self, func, *args, **kwargs):
            self.func = func
            super(FileChangeSensor.InotifyEventHandler, self).__init__(*args, **kwargs)

        def process_default(self, event):
            self.func()

    def notify(self):
        self.status += 1

    def setup(self):
        if self._notifier:
            self._notifier.stop()
        wm = pyinotify.WatchManager()
        handler = self.InotifyEventHandler(self.notify)
        self._notifier = pyinotify.ThreadedNotifier(wm, default_proc_fun=handler)

        wm.add_watch(self.filename, self.watch_flags, rec=True)
        self._notifier.start()

    def cleanup(self):
        self._notifier.stop()


class AbstractPollingSensor(AbstractSensor):

    """ Abstract baseclass for sensor that polls periodically its status"""

    #: How often to do polling
    interval = CFloat(5)

    #: This can be used to enable/disable polling
    poll_active = CBool(True)

    _stop = CBool(False, transient=True)
    _pollthread = Any(transient=True)
    view = AbstractSensor.view + ["interval"]
    silent = CBool(True)

    def setup(self):
        self._restart()

    def _poll_active_changed(self, old, new):
        if not self.traits_inited():
            return
        if new:
            self._restart()
        else:
            self._pollthread.cancel()

    def _restart(self):
        if self._stop:
            return
        if self._pollthread and self._pollthread.is_alive():
            self._pollthread.cancel()
        if self.poll_active:
            self.update_status()
            self._pollthread = threading.Timer(self.interval, threaded(self._restart))
            time_after_interval = datetime.now() + timedelta(seconds=self.interval)
            self._pollthread.name = "PollingSensor: %s next poll at %s (%.2f sek)" % (self.name, time_after_interval, self.interval)
            self._pollthread.start()

    def update_status(self):
        pass

    def cleanup(self):
        if self._pollthread:
            self._stop = True
            self._pollthread.cancel()


class PollingSensor(AbstractPollingSensor):

    """
        Polling sensor that uses a Callable when setting the status of the sensor.
    """

    #: Return value of this Callable is used to set the status of the sensor when polling
    status_updater = Instance(AbstractCallable)

    #: If set, typeconversion to this is used. Can be any function or type.
    type = Any

    callables = AbstractPollingSensor.callables + ['status_updater']

    def get_default_callables(self):
        from automate.callables import Empty
        c = super(PollingSensor, self).get_default_callables()
        c['status_updater'] = Empty()
        return c

    def setup(self):
        self.status_updater.setup_callable_system(self.system)
        self.on_trait_change(lambda: self.status_updater.setup_callable_system(self.system), 'status_updater')
        super(PollingSensor, self).setup()

    def update_status(self):
        if self.type is None:
            self.status = self.status_updater.call(self)
        else:
            self.status = self.type(self.status_updater.call(self))


class SimplePollingSensor(AbstractPollingSensor):

    """
        Polling sensor that calls on_update periodically.
    """

    def update_status(self):
        self.on_update.call(self)


class IntervalTimerSensor(AbstractPollingSensor):

    """
        Sensor that switches status between True and False periodically.
    """

    _status = CFloat

    def update_status(self):
        self.status = False if self.status else True


class SocketSensor(AbstractSensor):

    """
        Sensor that reads a TCP socket.

        Over TCP port, it reads data per lines and tries to set the status of the sensor
        to the value specified by the line. If content of the line is 'close', then connection
        is dropped.
    """

    #: Hostname/IP to listen. Use ``'0.0.0.0'``  to listen all interfaces.
    host = CStr('0.0.0.0')

    #: Port to listen
    port = CInt

    #: set to ``True`` to tell SocketSensor to stop listening to port
    stop = CBool(transient=True)

    _socket = Instance(socket.socket, transient=True)
    _status = CInt

    def listen_loop(self):
        while not self.stop:
            try:
                self.logger.info('%s listening to connections in port %s', self.name, self.port)
                self._socket.listen(1)
                self._socket.settimeout(1)
                while not self.stop:
                    try:
                        conn, addr = self._socket.accept()
                    except socket.timeout:
                        continue
                    break
                self.logger.info('%s connected from %s', self.name, addr)
                conn.settimeout(1)
                while not self.stop:
                    try:
                        data = conn.recv(1024)
                        if not data:
                            break
                        self.status = int(data.strip())
                        conn.sendall('OK\n')
                    except socket.timeout:
                        data = ''
                    except ValueError:
                        if data.strip() == 'close':
                            break
                        conn.sendall('NOK\n')
            except socket.error as e:
                self.logger.info("%s: Error %s caught.", self, e)
            except:
                if self.stop:
                    return
                else:
                    raise
            conn.close()
            self.logger.info('%s: connection %s closed', self.name, addr)
        self._socket.close()

    def setup(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.bind((self.host, self.port))
        t = threading.Thread(target=self.listen_loop, name='SocketSensor %s' % self.name)
        t.start()

    def cleanup(self):
        self.stop = True


class ShellSensor(AbstractSensor):

    """
        Run a shell command and follow its output. Status is set according to output, which is
        filtered through custome filter function.
    """

    #: Command can be, for example, 'tail -f logfile.log', which is convenient approach to follow log files.
    cmd = CStr
    #: If this is set to true, caller object is passed to the filter function as second argument
    caller = CBool
    filter = Any
    """
        Filter function, which must be a generator, such as for example:

        .. code-block:: python

            def filter(queue):
                while True:
                    line = queue.get()
                    if line == 'EOF':
                        break
                    yield line

        or a simple line-by-line filter::

            def filter(line):
                return processed(line)
    """

    _simple = CBool
    _stop = CBool
    _queue = Any(transient=True)
    _process = Any(transient=True)

    def cmd_loop(self):
        p = self._process = subprocess.Popen(self.cmd, shell=True, executable='bash', stdout=subprocess.PIPE)
        while True:
            line = p.stdout.readline().decode('utf-8')
            self._queue.put(line)
            if not line:
                self.logger.debug('Process exiting (cmd_loop)')
                break

    def status_loop(self):
        args = (self,) if self.caller else ()

        def default_filter(queue, *args):
            while True:
                line = queue.get()
                if not line:
                    self.logger.debug('Process exiting (status_loop)')
                    break
                yield line

        def simple_filter(line, *args):
            return line

        # Let's test if filter is 'simple' or not
        if self.filter:
            tst = self.filter('test line')
            if not isinstance(tst, types.GeneratorType):
                self._simple = True

        if self._simple:
            filter = self.filter or simple_filter
            while True:
                line = self._queue.get()
                if not line:
                    self.logger.debug('Process exiting (status_loop)')
                    break
                self.status = filter(line, *args)

        else:
            filter = self.filter or default_filter
            for s in filter(self._queue, *args):
                if self._stop:
                    break

                self.status = s

    def setup(self):
        self._queue = queue.Queue()
        t1 = threading.Thread(target=self.cmd_loop, name='ShellSensor.cmd_loop %s' % self.name)
        t1.start()
        t2 = threading.Thread(target=self.status_loop, name='ShellSensor.status_loop %s' % self.name)
        t2.start()
        self.logger.debug('Threads started')

    def cleanup(self):
        self._process.terminate()
        self.logger.debug('Process exiting (cleanup)')
        self._stop = True

__all__ = get_modules_all(AbstractSensor, locals())



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-webui.
#
# automate-webui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-webui is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-webui.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

import re
import operator
import logging

from django import template
from django.core.urlresolvers import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from automate.statusobject import StatusObject
from automate.program import Program

from ..forms import QUICK_EDITS

logger = logging.getLogger('automate.webui')

register = template.Library()


@register.simple_tag(takes_context=True)
def mocked_services(context):
    system = context.get('system')
    return [(service.__class__.__name__, service) for service in system.services if service.is_mocked]


@register.assignment_tag()
def ident(arg):
    return id(arg)


@register.assignment_tag()
def cancelable_thread(arg):
    return hasattr(arg, 'cancel')


@register.simple_tag
def program_status(actuator, program):
    try:
        return active_color((actuator, program), actuator.program_status[program])
    except KeyError:
        return '-'


@register.filter
def name_sort(l):
    return sorted(list(l), key=operator.attrgetter('class_name', 'name'))


def active_color(obj, value, changing=None, display=None, program=False):
    if isinstance(obj, tuple):
        id = u'program_status_for_actuator'
    else:
        id = u'object_status_' + obj.name
    if program:
        id = u'program_active_' + obj.name

    if display is None:
        display = value
    if changing:
        rv = u'<em class="status_changing %s">%s</em>' % (id, display)
    else:
        if value:
            rv = u'<em class="status_active %s">%s</em>' % (id, display)
        else:
            rv = u'<em class="status_inactive %s">%s</em>' % (id, display)

    return mark_safe(rv)


@register.simple_tag()
def program_active(obj):
    return active_color(obj, obj.active, False, program=True)


@register.simple_tag(takes_context=True)
def row_attrs(context, name, prefix=""):
    source = context['source']
    url = reverse('info_panel', args=(name,))
    return format_html('data-url="{url}?source={source}" data-name="{name}"',
        name=name,
        url=url,
        source=source,
    )


@register.simple_tag()
def object_status(obj):
    if isinstance(obj, Program):
        return active_color(obj, obj.active, False)
    if isinstance(obj, StatusObject):
        return active_color(obj, obj.status, obj.changing, display=obj.get_status_display())
    else:
        return active_color(obj, obj.get_status_display(), False)


@register.simple_tag
def condition_string(prog, attr):
    def repl(matchobj):
        if matchobj.group(1) == '__ACT__':
            return u'<em class="condition_active">%s</em>' % matchobj.group(2)
        else:
            return u'<em class="condition_inactive">%s</em>' % matchobj.group(2)

    cond = getattr(prog, attr)
    cond_str = cond.give_str_indented(tags=True)

    cond_str = re.sub(r"(__ACT__|__INACT__)('[^']*')", repl, cond_str)
    cond_str = re.sub(r"(__ACT__|__INACT__)(\"[^\"]*\")", repl, cond_str)
    cond_str = re.sub(r"(__ACT__|__INACT__)(\w*\()", repl, cond_str)
    cond_str = re.sub(r"(__ACT__|__INACT__)([\w=]*\b)", repl, cond_str)

    return mark_safe('<pre class="conditions">%s</pre>' % cond_str)


@register.assignment_tag(takes_context=True)
def sensor_form(context, sensor):
    try:
        form = QUICK_EDITS[sensor.data_type]({'status': sensor.status, 'name': sensor.name},
                                             source=context.get('source', 'main'), sensor=sensor)
    except Exception as e:
        logger.error('Error: %s', e)
        raise
    return form



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
import logging

from traits.api import HasStrictTraits, Instance, CBool

from automate.common import SystemBase


class AbstractService(HasStrictTraits):

    """
        Base class for System and UserServices
    """

    #: If set to *True*, service is loaded automatically (if not explicitly prevented
    #: in :attr:`automate.system.System.exclude_services`). Overwrite this in subclasses,
    autoload = False
    system = Instance(SystemBase)
    logger = Instance(logging.Logger, transient=True)
    is_mocked = CBool(False, transient=True)

    @property
    def id(self):
        return id(self)

    @property
    def name(self):
        return self.__class__.__name__

    def setup_system(self, system, name=None):
        self.system = system
        self.logger = self.system.logger.getChild(self.__class__.__name__)
        self.logger.info('Setup')
        self.system.register_service(self)
        self.setup()

    def cleanup_system(self):
        if self.system:
            self.logger.info('Cleaning up')
            self.cleanup()

    def setup(self):
        """
            Initialize service here. Define in subclasses.
        """

    def cleanup(self):
        """
            Cleanup actions must be performed here. This must be blocking until service is
            fully cleaned up.

            Define in subclasses.
        """

    def reload(self):
        self.cleanup()
        self.setup()
        self.logger.info('Reloading %s ready!', self)

    def __repr__(self):
        return '<' + self.__class__.__name__ + ' instance%s>' % ('' if self.system else ' not initialized')


class AbstractUserService(AbstractService):

    """Baseclass for UserServices. These are set up on startup. They provide usually user interaction services."""


class AbstractSystemService(AbstractService):

    """Baseclass for SystemServices. These are set up by when first requested by Sensor or other object."""

    autoload = True



from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import object
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

import queue
import logging
import threading


class StatusWorkerTask(object):

    def __init__(self, func, args, object):
        self.func = func
        self.args = args
        self.object = object

    @property
    def status(self):
        return self.args[0]

    def run(self):
        with self.object._status_lock:
            if self is self.object._queued_job:
                self.object.logger.debug('Running task %s', self)
                try:
                    self.func(*self.args)
                except Exception as e:
                    self.object.logger.error('Error running Task: %s', e)
                self.object._queued_job = None
            else:
                self.object.logger.debug('Task %s was not executed (was removed or replaced)', self)

    def __repr__(self):
        try:
            return '<Task %s %s>' % (self.object, self.args)
        except NameError:
            return '<Task *>'


class DummyStatusWorkerTask(object):

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def run(self):
        self.func(*self.args, **self.kwargs)

    def __repr__(self):
        return '<Dummy %s %s %s>' % (self.func, self.args, self.kwargs)


class StatusWorkerThread(threading.Thread):

    def _set_stop(self):
        self._stop = True
        self.logger.debug('Stop set')

    def __init__(self, *args, **kwargs):
        self.queue = queue.Queue()
        self._stop = False
        self.logger = logging.getLogger('automate.StatusWorkerThread')
        super(StatusWorkerThread, self).__init__(*args, **kwargs)

    def manual_flush(self):
        if self.is_alive():
            self.logger.error('Worker thread is running, cannot flush manually')
            return

        self.logger.info('Flushing queue manually (worker thread not yet alive)')
        while self.queue.queue:
            self.process_job()

    def process_job(self):
        job = self.queue.get()
        try:
            job.run()
        except Exception as e:
            self.logger.error('Error occurred when executing job %s: %s', job, e)
        self.queue.task_done()

    def run(self):
        self.logger.debug('StatusWorkerThread starting')
        while not self._stop:
            self.process_job()

        self.logger.debug('StatusWorkerThread exiting, entries: %s', self.queue.queue)

    def flush(self):
        """
        This only needs to be called manually
        from unit tests
        """
        self.logger.debug('Flush joining')
        self.queue.join()
        self.logger.debug('Flush joining ready')

    def put(self, job):
        self.logger.debug('Putting now %s', id(job))
        self.queue.put(job)

    def stop(self):
        self.logger.debug('Stopping: pre-flush')
        self.flush()

        self.put(DummyStatusWorkerTask(self._set_stop))

        self.logger.debug('Stopping')
        self.logger.debug('... entries: %s', self.queue.queue)
        self.flush()
        self.logger.debug('Stopping ready')



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-rpc.
#
# automate-rpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-rpc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-rpc.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from builtins import object
from . import wsgi_xmlrpc
from traits.api import CSet, Str, Any

from automate.extensions.wsgi.abstractwsgi import TornadoService


class ExternalApi(object):

    def __init__(self, system, tag):
        self.system = system
        self.tag = tag

    def set_status(self, name, status):
        """
            Set sensor ``name`` status to ``status``.
        """
        getattr(self.system, name).status = status
        return True

    def get_status(self, name):
        """
            Get status of object with name ``name``.
        """
        return getattr(self.system, name).status

    def set_object_status(self, statusdict):
        """
            Set statuses from a dictionary of format ``{name: status}``
        """
        for name, value in statusdict.items():
            getattr(self.system, name).status = value
        return True

    def toggle_object_status(self, objname):
        """
            Toggle boolean-valued sensor status between ``True`` and ``False``.
        """
        o = getattr(self.system, objname)
        o.status = not o.status
        self.system.flush()
        return o.status

    def get_sensors(self):
        """
            Get sensors as a dictionary of format ``{name: status}``
        """
        return {i.name: i.status for i in self.system.sensors}

    def get_websensors(self):
        """
            Get sensors with defined tag as a dictionary of format ``{name: status}``
        """
        return {i.name: i.status for i in self.system.sensors if self.tag & i.tags}

    def get_actuators(self):
        """
            Get actuators as a dictionary of format ``{name: status}``
        """
        return {i.name: i.status for i in self.system.actuators}

    def flush(self):
        """
            Flush the system queue. If you have just set a status and then read a value,
            it might be necessary to flush queue first, such that related changes have been
            applied.
        """
        self.system.flush()

    def is_alive(self):
        """
            Simple RPC command that returns always True.
        """
        return True

    def log(self):
        """
            Return recent log entries as a string.
        """
        logserv = self.system.request_service('LogStoreService')
        return logserv.lastlog(html=False)


class RpcService(TornadoService):

    #: Tags that are displayed via get_websensors RPC function
    view_tags = CSet(trait=Str, value={'rpc'})

    #: If you want to define custom api (similar to, or derived from :class:`.ExternalApi`, it can be given here.
    api = Any

    def get_wsgi_application(self):
        instance = self.api or ExternalApi(self.system, tag=self.view_tags)
        wsgiapp = wsgi_xmlrpc.WSGIXMLRPCApplication(instance=instance)
        return wsgiapp



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-webui.
#
# automate-webui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-webui is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-webui.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

"""
    Web interface module for Automate
"""
from __future__ import unicode_literals

import json
import datetime
import time
import os

from django.template.loader import render_to_string
import tornado.web
from tornado.websocket import WebSocketHandler, WebSocketClosedError

from traits.api import CBool, Tuple, Int, Str, CSet, List, CInt, Dict, Unicode

from automate.extensions.webui.djangoapp.views import set_globals
from automate.statusobject import StatusObject
from django.core.wsgi import get_wsgi_application
from automate.extensions.wsgi import TornadoService


class WebService(TornadoService):

    """
    Web User Interface Service for Automate
    """

    #: Restrict usage to only monitoring statuses (default: ``True``).
    #: If WebService is not in read_only mode, it is possible to run arbitrary Python commands
    #: through eval/exec via web browser. This is, of course, a severe security issue.
    #: Secure SSL configuration HIGHLY recommended, if not operating in ``read_only`` mode.
    read_only = CBool(True)

    #: Default view that is displayed when entering the server. Can be the name of any view in views.py
    default_view = Str('system')

    #: Below Actuator row, show active Programs that are controlling Actuator
    show_actuator_details = CBool(True)

    #: HTTP port to listen
    http_port = Int(8080)

    #: Authentication for logging into web server. (user,password) pairs in a tuple.
    http_auth = Tuple

    #: Let websocket connection die after ``websocket_timeout`` time of no ping reply from client.
    websocket_timeout = CInt(60 * 5)

    #: Tags that are shown in user defined view
    user_tags = CSet(trait=Str, value={'user'})

    #: Django debugging mode (slower, more info shown when error occurs)
    debug = CBool(False)

    #: User-defined custom pages as a dictionary of form ``{name: template_content}``
    custom_pages = Dict(key_trait=Unicode, value_trait=Unicode)

    #: set to True, if you want to launch multiple servers with same system. Authentication and
    #: other settings are then taken from master service. Only web server settings (http host/port)
    #: are used from slave.
    slave = CBool(False)

    #: In this dictionary you can define your custom Django settings which will override the default ones
    django_settings = Dict()

    #: From /set/object/value and /toggle/object, redirect to /set_ready/object/value after after executing action
    redirect_from_setters = CBool(True)

    #: Show hidden objects in web UI (those prefixed with _)
    show_hidden = CBool(False)

    #: Full exposed url root
    server_url = Unicode()

    _sockets = List

    def get_filehandler_class(service):
        class MyFileHandler(tornado.web.StaticFileHandler):

            def validate_absolute_path(self, *args, **kwargs):
                session_id = getattr(self.request.cookies.get('sessionid', None), 'value', None)
                from django.contrib.sessions.middleware import SessionMiddleware
                mw = SessionMiddleware()
                session_data = mw.SessionStore(session_id)
                if not session_data.get('logged_in', False):
                    raise tornado.web.HTTPError(403, 'not logged in')

                return super(MyFileHandler, self).validate_absolute_path(*args, **kwargs)

            def check_etag_header(self):
                """ Disable etag_header checking (checks only modified time). Due to etag caching
                    file changes were not detected at all. """
                return False
        return MyFileHandler

    def get_tornado_handlers(self):
        if self.slave:
            return self.system.request_service('WebService').get_tornado_handlers()
        super_handlers = super(WebService, self).get_tornado_handlers()
        path = os.path.join(os.path.dirname(__file__), 'static')
        static = [('/static/(.*)', tornado.web.StaticFileHandler, {'path': path})]
        return static + super_handlers

    def setup(self):
        if not self.slave:
            os.environ['DJANGO_SETTINGS_MODULE'] = 'automate.extensions.webui.settings'
            from django.conf import settings
            settings.DEBUG = self.debug
            if not 'SECRET_KEY' in self.django_settings:
                self.logger.warning('Insecure settings! Please set proper SECRET_KEY in django_settings!')
            if not 'TIME_ZONE' in self.django_settings:
                os.environ.pop('TZ')  # Django uses America/Chicago as default timezone. Let's clean this up.
                time.tzset()
            if self.server_url:
                self.django_settings['SERVER_URL'] = self.server_url
            for key, value in self.django_settings.items():
                setattr(settings, key, value)

            set_globals(self, self.system)

        super(WebService, self).setup()
        if not self.slave:
            self.system.request_service('LogStoreService').on_trait_change(self.push_log, 'most_recent_line')

            self.system.on_trait_change(self.update_sockets, 'objects.status, objects.changing, objects.active, '
                                        'objects.program_status_items')

    def get_websocket(service):
        if service.slave:
            return service.system.request_service('WebService').get_websocket()

        class WebSocket(WebSocketHandler):

            def data_received(self, chunk):
                pass

            def __init__(self, application, request, **kwargs):
                self.log_requested = False
                self.subscribed_objects = set()
                self.last_message = None
                self.logged_in = False

                super(WebSocket, self).__init__(application, request, **kwargs)

            def check_origin(self, origin):
                return True

            def write_json(self, **kwargs):
                msg = json.dumps(kwargs)
                service.logger.debug('Sending to client %s', msg)
                self.write_message(msg)

            def open(self):
                self.session_id = session_id = getattr(self.request.cookies.get('sessionid', None), 'value', None)
                from django.contrib.sessions.middleware import SessionMiddleware
                mw = SessionMiddleware()
                session_data = mw.SessionStore(session_id)

                if session_data.get('logged_in', False) or not service.http_auth:
                    self.logged_in = True
                else:
                    service.logger.warning("Not (yet) logged in %s", session_id)

                service.logger.debug("WebSocket opened for session %s", session_id)
                service._sockets.append(self)

            def _authenticate(self, username, password):
                if (username, password) == service.http_auth:
                    self.logged_in = True
                    service.logger.debug('Websocket user %s logged in.', username)
                else:
                    service.logger.warning('Authentication failure: user %s with passwd %s != %s ',
                                           username, password, service.http_auth)
                    self.close()

            def _ping(self):
                pass

            def _set_status(self, name, status):
                if service.read_only:
                    service.logger.warning("Could not perform operation: read only mode enabled")
                    return
                obj = service.system.namespace.get(name, None)
                if obj:
                    obj.status = status

            def _subscribe(self, objects):
                self.subscribed_objects.update(objects)

            def _unsubscribe(self, objects):
                self.subscribed_objects -= set(objects)

            def _clear_subscriptions(self):
                self.subscribed_objects.clear()

            def _send_command(self, command):
                if not service.read_only:
                    service.system.cmd_exec(command)
                else:
                    service.logger.warning("Could not perform operation: read only mode enabled")

            def _fetch_objects(self):
                data = [(i.name, i.get_as_datadict()) for i in service.system.objects_sorted]
                self.write_json(action='return', rv=data)

            def _request_log(self):
                self.log_requested = True

            def on_message(self, json_message):
                """
                    Message format: {'action': 'action_name', other_kwargs...)
                """

                message = json.loads(json_message)
                service.logger.debug('Message received from client: %s', message)

                action = message.pop('action', '')

                if self.logged_in:
                    action_func = getattr(self, '_' + action, None)
                    if action_func:
                        service.logger.debug('Running websocket action %s', action)
                        action_func(**message)
                    else:
                        service.logger.error('Not logged in or unknown message %s', message)
                elif action == 'authenticate':
                    return self._authenticate(**message)

                self.last_message = datetime.datetime.now()

            def on_close(self):
                service.logger.debug("WebSocket closed for session %s", self.session_id)
                service._sockets.remove(self)

        return WebSocket

    def push_log(self, new):
        for s in self._sockets:
            if s.log_requested:
                try:
                    s.write_json(action='log', data=new)
                except WebSocketClosedError:
                    pass

    def update_sockets(self, obj, attribute, old, new):
        if isinstance(obj, StatusObject):
            self.logger.debug('Update_sockets %s %s %s %s', obj, attribute, old, new)
            for s in self._sockets:
                if s.last_message and (s.last_message < datetime.datetime.now()
                                       - datetime.timedelta(seconds=self.websocket_timeout)):
                    self.logger.info('Closing connection %s due to timeout', s.session_id)
                    s.on_close()
                    s.close(code=1000, reason='Timeout')
                    continue
                if obj.name in s.subscribed_objects:
                    if attribute == 'program_status_items' and self.show_actuator_details:
                        new_actuator_html = render_to_string(
                            'rows/actuator_row.html', dict(actuator=obj, source='__SOURCE__', service=self))
                        s.write_json(action='update_actuator', name=obj.name, html=new_actuator_html)
                    elif attribute == 'active':
                        s.write_json(action='program_active', name=obj.name, active=obj.active)
                    elif attribute in ['status', 'changing']:
                        s.write_json(action='object_status',
                                     name=obj.name,
                                     status=obj.status,
                                     display=obj.get_status_display(),
                                     changing=obj.changing)

    def get_wsgi_application(self):
        return get_wsgi_application()



# (c) 2017 Tuomas Airaksinen
# 
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from django.core.urlresolvers import reverse
from django.test import Client
from future.backports.urllib.parse import urlparse

from automate import *
from automate.extensions.webui import WebService


@pytest.fixture()
def sys_with_web():
    web = WebService(
        http_port=80811,
        http_auth=(
            ('test', 'test')
        ),
        debug=False,
        user_tags={'web'},
        read_only=False,
        default_view='tags_view',
        show_actuator_details=False,
        #static_dirs={'/webcam/(.*)': '/home/tuma/public_html/webcam/'},
        #custom_pages={'Webcam': webcam_page},
        #django_settings={
        #    'SESSION_FILE_PATH': '/var/cache/automatesession' if is_raspi() else '/tmp',
        #    'SESSION_COOKIE_AGE': 52560000},
    )
    class sys(System):
        s1 = UserBoolSensor()
        s2 = UserBoolSensor()
        a1 = BoolActuator(default=False)
        a2 = BoolActuator()

    s = sys(exclude_services=['TextUIService'], services=[web])
    s.flush()
    yield s
    s.cleanup()

class Http:
    OK = 200
    REDIR = 302
    FORBIDDEN = 304


def r(view_name, *args, **kwargs):
    return reverse(view_name, args=args, kwargs=kwargs)

@pytest.fixture
def constants(sys_with_web):
    class Pages:
        ROOT = '/'
        LOGIN = r('login')

        MAIN = r('main')
        TAGS = r('tags_view')
        USER1 = r('user_editable_view')
        USER2 = r('user_defined_view')
        GROUPS = r('tag_view_only_groups')
        CONSOLE = r('console')
        NEW_ACTUATOR = r('new', 'actuator')
        NEW_SENSOR = r('new', 'sensor')
        NEW_PROGRAM = r('new', 'program')

        BASIC_VIEWS = [TAGS, USER1, USER2, GROUPS, CONSOLE, NEW_ACTUATOR, NEW_PROGRAM, NEW_SENSOR]

        LOGOUT = r('logout')

    return Pages


@pytest.fixture()
def logged_client(sys_with_web, constants):
    client = Client()
    res = client.get('/')
    assert res.status_code == Http.REDIR
    assert u(res.url).startswith('/login')
    # let's login
    res = client.post(u(res.url), {'username': 'test', 'password': 'test'})
    assert res.status_code == Http.REDIR
    assert u(res.url) == constants.ROOT
    res = client.get(constants.ROOT)
    assert res.status_code == Http.REDIR
    assert u(res.url) == constants.TAGS

    yield client

    res = client.get(constants.LOGOUT)
    for p in constants.BASIC_VIEWS:
        res = client.get(p)
        assert res.status_code == Http.REDIR
        assert u(res.url).startswith(constants.LOGIN)


def u(url):
    return urlparse(url).path

def test_web(sys_with_web, logged_client, constants):
    for p in constants.BASIC_VIEWS:
        res = logged_client.get(p)
        assert res.status_code == Http.OK


# TODO:
# - test creating new objects
# - test editing objects via web
# - test changing status via web, via REST AND via websockets
# - test that console gets input
# - test making change via console




# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

"""
Apply two fixeds to traits 4.5.0 that are necessary in order to Automate to function.
Most likely these are not necessary (neither harmful) any more with 4.6.0.
"""
from __future__ import unicode_literals

import traits.traits_listener as tlistener
original_register = tlistener.ListenerItem.register


def ListenerItem_register(self, new):
    import traits.has_traits as has_traits
    if not isinstance(new, has_traits.HasTraits):
        return tlistener.INVALID_DESTINATION

    return original_register(self, new)
tlistener.ListenerItem.register = ListenerItem_register
import traits.trait_types as ttypes


def CSet_validate(self, object, name, value):
    if not isinstance(value, set):
        try:
            value = set(value)
        except (ValueError, TypeError):
            value = {value}

    return super(ttypes.CSet, self).validate(object, name, value)

ttypes.CSet.validate = CSet_validate

# This hack is necessary in order to silence undesired error messages from Traits notification handler

import traits.trait_notifiers as tnotifier


def _dispatch_change_event(self, object, trait_name, old, new, handler):
    """ Prepare and dispatch a trait change event to a listener. """

    # Extract the arguments needed from the handler.
    args = self.argument_transform(object, trait_name, old, new)

    # Send a description of the event to the change event tracer.
    if tnotifier._pre_change_event_tracer is not None:
        tnotifier._pre_change_event_tracer(object, trait_name, old, new, handler)

    # Dispatch the event to the listener.
    from automate.common import SystemNotReady
    try:
        self.dispatch(handler, *args)
    except SystemNotReady:
        pass
    except Exception as e:
        if tnotifier._post_change_event_tracer is not None:
            tnotifier._post_change_event_tracer(object, trait_name, old, new,
                                                handler, exception=e)
        # This call needs to be made inside the `except` block in case
        # the handler wants to re-raise the exception.
        tnotifier.handle_exception(object, trait_name, old, new)
    else:
        if tnotifier._post_change_event_tracer is not None:
            tnotifier._post_change_event_tracer(object, trait_name, old, new,
                                                handler, exception=None)

tnotifier.TraitChangeNotifyWrapper._dispatch_change_event = _dispatch_change_event



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-rpio.
#
# automate-rpio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-rpio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-rpio.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from .rpio_actuators import RpioActuator, RpioPWMActuator
from .rpio_sensors import RpioSensor, TemperatureSensor
from .rpio_service import RpioService

extension_classes = [RpioActuator, RpioSensor, RpioService, TemperatureSensor, RpioPWMActuator]



from __future__ import unicode_literals
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from automate.systemobject import SystemObject
from automate.common import Group
from automate.service import AbstractService
from automate.callable import AbstractCallable


class Namespace(dict):
    """
        Namespace dictionary object for Automate System.
        Contains all SystemObjects stored in System by their names.
    """

    def __init__(self, system=None, *args, **kwargs):
        self.allow_overwrite = []
        self['system'] = self.system = system
        self['reverse'] = self.reverse = {}
        self['logger'] = self.logger = system.logger.getChild('logger')
        self['__name__'] = self.system.name + '_namespace'
        super(Namespace, self).__init__(*args, **kwargs)

    def give_systemobjects(self, system, tags=None):
        objs = []
        if not tags:
            tags = set()
        import inspect
        for name, obj in inspect.getmembers(system):
            if isinstance(obj, SystemObject):
                objs.append((name, obj, tags))
            if isinstance(obj, type) and issubclass(obj, Group):
                add_tags = tags.copy()
                if hasattr(obj, 'tags'):
                    add_tags = add_tags | set(obj.tags.split(','))
                objs.extend(self.give_systemobjects(obj, add_tags | {'group:%s' % name}))
        return objs

    def set_system(self, loadstate=None):
        if loadstate:
            objs = [(i._passed_arguments[1]['name'], i, []) for i in loadstate]
        else:
            objs = self.give_systemobjects(self.system)

        def order(x):
            if isinstance(x[1], AbstractCallable):
                # first handle plain callables
                ord = float('-inf')
            else:
                ord = x[1]._order
            return ord

        objs.sort(key=order)
        if objs:
            SystemObject._count = max(SystemObject._count, objs[-1][1]._order + 1)

        self.system.logger.info('Setup obj.system and names in namespace')

        for name, obj, groups in objs:
            obj.system = self.system
            if name in self:
                raise NameError('%s already in namespace!' % name)
            self[name] = obj

        self.logger.info('Set up system and groups into object tags')
        for name, obj, groups in objs:
            obj.setup_system(self.system, name, loadstate=loadstate)

            if not loadstate:
                is_groups = False
                for g in groups:
                    obj.tags.add(g)
                    if g.startswith('group:'):
                        is_groups = True
                if not is_groups:
                    obj.tags.add('group:root')

        # flush, so that sensor default initial statuses are up to date
        self.system.worker_thread.manual_flush()

        self.system.logger.info('Setup callables. This activates program features.')

        def order(x):
            if isinstance(x[1], AbstractCallable):
                # first handle plain callables
                ord = float('inf')
            else:
                # then programs starting from highest priority
                ord = getattr(x[1], 'priority', -1)
            return ord

        objs.sort(key=order, reverse=True)

        for name, obj, groups in objs:
            obj.setup_callables()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, super(Namespace, self).__repr__())

    def __delitem__(self, key):
        o = self[key]
        try:
            if o in self.reverse:
                del self.reverse[o]
        except TypeError as e:
            if not str(e).startswith("unhashable type"):
                raise

        if isinstance(o, SystemObject) and o in self.system.objects:
            self.system.objects.remove(o)
        super(Namespace, self).__delitem__(key)

    def update(self, d):
        for key, value in list(d.items()):
            self[key] = value

    def __setitem__(self, name, value):
        is_alias = False
        if name in self and name not in self.allow_overwrite:
            if isinstance(self[name], SystemObject):
                raise ValueError('Cannot overwrite %s in AutomateNamespace of %s' % (name, self.system.name))
            self.logger.warning('Overwriting %s in AutomateNamespace %s', name, self.system.name)

        try:
            if value in self.get('reverse', []):
                is_alias = True
        except TypeError as e:
            if not str(e).startswith("unhashable type"):
                raise

        super(Namespace, self).__setitem__(name, value)

        if name in self.allow_overwrite or is_alias:
            return
        if isinstance(value, AbstractService):
            self.system.services.append(value)
        elif isinstance(value, SystemObject):
            self.reverse[value] = name
            value.name_changed_event = True
            self.system.objects.add(value)

            if not value.system:
                value.setup_system(self.system, name)
                value.setup_callables()



# (c) 2015-2017 Tuomas Airaksinen
# 
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
import os
SECRET_KEY = 'insecure-default'
ALLOWED_HOSTS = ['*']
ROOT_URLCONF = 'automate.extensions.webui.djangoapp.urls'
USE_TZ=False

TEMPLATES = [
                {
                    'BACKEND': 'django.template.backends.django.DjangoTemplates',
                    'DIRS': [
                        # insert your TEMPLATE_DIRS here
                    ],
                    'APP_DIRS': True,
                    'OPTIONS': {
                        'context_processors': [
                            'django.template.context_processors.debug',
                            'django.template.context_processors.i18n',
                            'django.template.context_processors.media',
                            'django.template.context_processors.static',
                            'django.template.context_processors.tz',
                            'django.contrib.messages.context_processors.messages',
                            'automate.extensions.webui.djangoapp.views.common_context',
                        ],
                    },
                },
            ]

INSTALLED_APPS = ['crispy_forms',
                  'django.contrib.staticfiles',
                  'django.contrib.messages',
                  'automate.extensions.webui.djangoapp',
                  ]

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(os.path.dirname(__file__), 'static')
MESSAGE_STORAGE = 'django.contrib.messages.storage.session.SessionStorage'
MIDDLEWARE_CLASSES = [
#    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]
#STATICFILES_STORAGE = 'whitenoise.storage.CompressedManifestStaticFilesStorage'

SESSION_ENGINE = "django.contrib.sessions.backends.file"
CRISPY_TEMPLATE_PACK = 'bootstrap3'
LOGGING = {  # set up logging such that log entries go to Automate logging
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        }
    },
    'loggers': {
        'django': {
            'level': 'DEBUG',
            'handlers': ['null'],
        }
    }
}


from __future__ import unicode_literals
from builtins import object
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

import pytest

from automate import *
from automate.program import Program
from automate.statusobject import AbstractSensor, AbstractActuator

import time


def test_namespace():
    sys = System(exclude_services=['TextUIService'])
    ns = sys.namespace  # AutomateNamespace(sys)
    # ns.set_system()
    s = AbstractSensor()

    ns['sens'] = s
    assert s in ns.reverse
    assert 'sens' in ns
    #assert ns.sens is s
    assert s in sys.sensors

    ns['one'] = 1
    assert 1 not in ns.reverse

    a = AbstractActuator()
    ns['act'] = a
    with pytest.raises(ValueError):
        ns['act'] = a

    assert a in sys.actuators
    assert a not in sys.sensors

    p = Program()
    ns['prg'] = p
    assert p in sys.programs
    #assert 'sensors' in ns
    #assert 'programs' in ns
    #assert 'actuators' in ns
    del ns['prg']
    assert not p in sys.programs
    del ns['act']
    assert not a in sys.actuators
    sys.cleanup()


@pytest.fixture(params=[0, 1])
def mysys(request):
    class sysclass(System):
        mysensor = UserFloatSensor()
        myactuator = FloatActuator()
        prog = Program(active_condition=Value(mysensor if request.param else 'mysensor'),
                       on_update=Run(
                           SetStatus(myactuator if request.param else 'myactuator', 1),
                           Debug("Hello World!")),
                       )
    s = sysclass(exclude_services=['TextUIService'])

    yield s
    s.cleanup()


def test_tags(mysys):
    mysys.mysensor.tags = ['hep', 'hop']
    assert mysys.mysensor.tags == {'hep', 'hop'}
    mysys.mysensor.tags = 'hup,hip'
    assert mysys.mysensor.tags == {'hup', 'hip'}

    #['hep', 'hop']


def test_logging(caplog, mysys):
    out = caplog.text()
    s = 'Initializing services'
    assert s in out
    out = out.replace(s, '')
    assert s not in out


def test_automatesystem(mysys):
    s = mysys
    assert s.mysensor.status == False
    assert s.myactuator.status == False
    s.mysensor.set_status(1)
    s.flush()
    assert s.myactuator.status == True
    ns = s.namespace


def test_register_function(mysys):
    def func():
        return True
    mysys.register_service_functions(func)
    assert mysys.namespace['func']()
    assert mysys.func()

    class tst(object):

        def hep(self):
            return self.status
    t = tst()
    t.status = True
    mysys.register_service_functions(t.hep)
    assert mysys.hep()


def test_namechange(mysys):
    assert mysys.mysensor.name == 'mysensor'
    mysys.mysensor.name = 'newname'
    assert mysys.mysensor.name == 'newname'
    assert 'newname' in mysys.namespace
    assert 'mysensor' not in mysys.namespace


def test_new_systemobj(mysys):
    a1 = FloatActuator('somename', system=mysys)
    assert 'somename' in mysys.namespace
    mysys.prog.on_update = Run(SetStatus(mysys.myactuator, 1), SetStatus(a1, 2.0))
    mysys.mysensor.set_status(1)
    mysys.flush()
    assert a1.status == 2.0
    a2 = FloatActuator('somename', system=mysys)
    assert a2.name == 'somename_00'
    a3 = FloatActuator('somename', system=mysys)
    assert a3.name == 'somename_01'
    a4 = FloatActuator(system=mysys)
    assert a4.name == 'Nameless_FloatActuator'

from automate.common import Lock


def test_lock(sysloader):
    # TODO: check why this is not robust (see https://travis-ci.org/tuomas2/automate/jobs/245711227)
    class mysys(System):
        prg = UserIntSensor()
    s = sysloader.new_system(mysys)

    l = Lock('name')
    flag = []

    def locktst():
        assert l.context
        l.release()
        flag.append(1)
    c = s.prg.on_deactivate = Delay(0.2, Func(locktst))
    c.call(s.prg)
    l.acquire()
    # now l is locked, but locktst will release it.
    l.acquire() # This will wait until l is released.

    assert flag
    l.release() # final release.


def test_sysobject_callable(sysloader):
    class mysys(System):
        mysens = UserIntSensor(on_activate=Run('mycal'), priority=50)
        mycal = Run(Run(Value(1), Shell('false')), Value(2))
    s = sysloader.new_system(mysys)


def test_sysobject_callable2(sysloader):
    class mysys(System):
        mysens = UserIntSensor(on_activate=Run('mycal'), priority=50)
        mycal = SetStatus('mysens', Add(Value('mysens'), 1))
    s = sysloader.new_system(mysys)

# test some Sensors


def test_shellsensor_defaultfilter(sysloader):
    class mysys(System):
        p1 = ShellSensor(cmd='echo test\necho test2\necho test3')
    s = sysloader.new_system(mysys)
    time.sleep(0.1)
    assert s.p1.status == 'test3\n'


def test_shellsensor_defaultfilter_simple(sysloader):
    class mysys(System):
        p1 = ShellSensor(cmd='echo test\necho test2\necho test3')
    s = sysloader.new_system(mysys)
    time.sleep(0.1)
    assert s.p1.status == 'test3\n'


def test_shellsensor_simple(sysloader):
    lines = []

    def myfunc(line):
        lines.append(line)
        return line[:-1]

    class mysys(System):
        p1 = ShellSensor(cmd='echo test\necho test2\necho test3', filter=myfunc)
    s = sysloader.new_system(mysys)
    time.sleep(0.1)
    assert s.p1.status == 'test3'
    assert lines == ['test line', 'test\n', 'test2\n', 'test3\n']


def test_shellsensor(sysloader):
    lines = []

    def myfunc(q):
        while True:
            line = q.get()
            if not line:
                break
            lines.append(line)
            yield line[:-1]

    class mysys(System):
        p1 = ShellSensor(cmd='echo test\necho test2\necho test3', filter=myfunc)
    s = sysloader.new_system(mysys)
    time.sleep(0.1)
    assert s.p1.status == 'test3'
    assert lines == ['test\n', 'test2\n', 'test3\n']


def test_shellsensor_delayed(sysloader):
    lines = []

    def myfunc(q):
        while True:
            line = q.get()
            if not line:
                break
            lines.append(line)
            yield line[:-1]

    class mysys(System):
        p1 = ShellSensor(cmd='echo;echo test\necho test2\necho\n\n\nsleep 0.1\necho test3', filter=myfunc)
    s = sysloader.new_system(mysys)
    time.sleep(1)
    assert s.p1.status == 'test3'
    assert lines == ['\n', 'test\n', 'test2\n', '\n', 'test3\n']


def test_setup_system(sysloader):
    class mysys(System):
        lamppu1 = BoolActuator()
        lamppu2 = BoolActuator()
        lamppu3 = BoolActuator()
        lamp_on_delay = UserFloatSensor()
        lamp_off_delay = UserFloatSensor()

        lamput = BoolActuator(
            active_condition=Value(True),
            on_update=Run(
                SetStatus(lamppu1, 'lamput'),
                Delay(IfElse('lamput', lamp_on_delay, lamp_off_delay),
                      SetStatus(lamppu3, 'lamput')),
                Delay(IfElse('lamput', Value(2) * lamp_on_delay, Value(2) * lamp_off_delay),
                      If(Not('lamput'),
                         Empty(),  # RemoteFunc(raspi2host, 'set_status', 'akvadimmer', 1)
                         ),
                      Delay(5,
                            SetStatus(lamppu2, 'lamput')),
                      )),
        )
    s = sysloader.new_system(mysys)
    a = s.lamput.on_update
    assert a.system  # run
    assert a[0].system  # setstatus
    assert a[1].system  # delay
    assert a[1][0].system  # ifelse
    assert a[1][1].system  # setstatus
    assert a[2].system  # delay
    assert a[2][0].system  # ifelse
    assert a[2][0][1].system  # mult
    assert a[2][0][1][0].system  # value
    assert a[2][0][2].system  # mult
    assert a[2][0][2][0].system  # value

#@mock.patch('traits_enaml.imports')
#@mock.patch('enaml.qt')
# def test_guithread(mock_qt, mock_traits):
#    g = GuiThread()
#    g.start()
#    g.stop = True
#    g.trigger()



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-rpio.
#
# automate-rpio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-rpio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-rpio.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division
from __future__ import unicode_literals
from traits.api import Instance, Int, Bool, Enum, CUnicode, CFloat, CBool
from automate.sensors import UserBoolSensor, AbstractPollingSensor, UserFloatSensor
from automate.service import AbstractSystemService


class RpioSensor(UserBoolSensor):

    """
        Boolean-valued sensor object that reads Raspberry Pi GPIO input pins.
    """

    user_editable = CBool(False)

    #: GPIO port
    port = Int

    #: Set to True to have inversed status value
    inverted = Bool(False)

    #: Button setup: "down": pushdown resistor, "up": pushup resistor, or "none": no resistor set up.
    button_type = Enum("down", "up", "none")

    view = UserBoolSensor.view + ["port", "button_type"]

    _hw_service = Instance(AbstractSystemService, transient=True)

    def setup(self):
        self._hw_service = self.system.request_service('RpioService')

        self._hw_service.enable_input_port(self.port, self.gpio_callback, self.button_type)

    def _button_type_changed(self, new):
        if self._hw_service:
            self._hw_service.disable_input_port(self.port)
            self._hw_service.enable_input_port(self.port, self.gpio_callback, new)

    def gpio_callback(self, gpio_id, value):
        self.set_status(value if not self.inverted else not value)

    def update_status(self):
        self.gpio_callback(None, self._hw_service.get_input_status(self.port))

    def _port_changed(self, old, new):
        if not self._hw_service:
            return
        if old:
            self._hw_service.disable_input_port(old)
        self._hw_service.enable_input_port(new, self.gpio_callback, self.button_type)


class TemperatureSensor(AbstractPollingSensor):

    """
        W1 interface (on Raspberry Pi board) that polls polling temperature.
        (kernel modules w1-gpio and w1-therm required).
        Not using RPIO, but placed this here, since this is also Raspberry Pi related sensor.
    """

    _status = CFloat

    #: Address of W1 temperature sensor (something like ``"28-00000558263c"``), see what you have in
    #: ``/sys/bus/w1/devices/``
    addr = CUnicode
    view = list(set(UserFloatSensor.view + AbstractPollingSensor.view + ["addr"]))

    #: Maximum jump in temperature, between measurements. These temperature sensors
    #: tend to give sometimes erroneous results.
    max_jump = CFloat(5.0)

    _error_count = Int(0, transient=True)

    #: Maximum number of erroneous measurements, until value is really set
    max_errors = Int(5)

    _first_reading = CBool(True, transient=True)

    def get_status_display(self, **kwargs):
        if 'value' in kwargs:
            value = kwargs['value']
        else:
            value = self.status
        return u"%.1f ⁰C" % value

    def update_status(self):
        w1file = "/sys/bus/w1/devices/%s/w1_slave" % self.addr
        try:
            f = open(w1file)
        except IOError:
            return

        try:
            temp = float(f.read().split("\n")[1].split(" ")[9].split("=")[1]) / 1000.
        except IOError:
            self.logger.error("IO-error in temperature sensor %s, not set", self.name)
            return
        if abs(temp-self.status) > self.max_jump and self._error_count < self.max_errors and not self._first_reading:
            self._error_count += 1
        else:
            self._first_reading = False
            self._error_count = 0
            self.set_status(temp)
        f.close()



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import absolute_import
from __future__ import unicode_literals

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("automate").version
except pkg_resources.DistributionNotFound:
    __version__ = 'not installed'

from . import traits_fixes
from .common import *
from .program import *
from .system import *
from .callables import *
from .actuators import *
from .sensors import *
from .services import *
from .tools import *
from .statusobject import *



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division

from builtins import range
from builtins import object
import pytest
import mock

from automate import *
from automate.program import Program
from automate.statusobject import AbstractSensor
import time


ORIGVAL = 1.25
NEWVAL = 2.0


@pytest.fixture(params=[0, 1])
def mysys(request):
    if request.param == 0:
        class _mysys(System):
            act = FloatActuator(default=ORIGVAL)
            a2 = FloatActuator(default=ORIGVAL)
            sens = UserBoolSensor(default=False)
            prog = Program(
                active_condition=Value(sens),
                on_activate=SetStatus(act, NEWVAL),
            )
    else:
        class _mysys(System):
            act = FloatActuator(default=ORIGVAL)
            a2 = FloatActuator(default=ORIGVAL)
            sens = UserBoolSensor(default=False)
            prog = Program(
                active_condition=Value('sens'),
                on_activate=SetStatus('act', NEWVAL),
            )
    sys = _mysys(exclude_services=['TextUIService'])
    sys.flush()
    yield sys
    sys.cleanup()


@pytest.fixture
def prog(request):
    class _mysys(System):
        act = FloatActuator(default=ORIGVAL)
        a2 = FloatActuator(default=ORIGVAL)
        sens = UserBoolSensor(default=False)
        prog = Program(
            active_condition=Value(sens),
            on_activate=SetStatus(act, NEWVAL),
        )
    sys = _mysys(exclude_services=['TextUIService'])
    sys.flush()
    yield sys.prog
    sys.cleanup()


def test_oftypes(sysloader):
    class _mysys(System):
        act = FloatActuator(default=ORIGVAL)
        a2 = FloatActuator(default=ORIGVAL)
        sens = UserBoolSensor(default=False)
        prog = Program(
            active_condition=OfType(FloatActuator),
        )
        prog2 = Program(
            active_condition=And(sens, 'all_floatactuators'),
        )
        prog3 = Program(
            active_condition=And(sens, OfType(FloatActuator, exclude=[a2])),
        )
        prog4 = Program(
            priority=2.0,
            active_condition=Value(sens),
            on_activate=SetStatus(OfType(FloatActuator, type='targets'), 3.0),
        )
        all_floatactuators = OfType(AbstractActuator)
        prog5 = Program(
            on_activate=Run(SetStatus(all_floatactuators, all_floatactuators)),
        )

    sys = sysloader.new_system(_mysys)
    assert sys.prog3.actual_triggers == {sys.act, sys.sens}
    assert sys.prog2.actual_triggers == {sys.act, sys.a2, sys.sens}
    assert sys.prog.actual_triggers == {sys.act, sys.a2}
    sys.prog.active_condition = OfType(FloatActuator, UserBoolSensor)
    assert sys.prog.actual_triggers == {sys.act, sys.a2, sys.sens}
    assert sys.prog4.actual_targets == {sys.act, sys.a2}
    assert sys.prog4.actual_triggers == {sys.sens}
    sys.sens.status = 1
    sys.flush()
    assert sys.prog4 in sys.act.program_stack
    assert sys.prog4 in sys.a2.program_stack
    assert sys.prog4 == sys.act.program
    assert sys.prog4 == sys.a2.program
    assert sys.act.status == 3.0
    assert sys.a2.status == 3.0
    sys.prog.active_condition = OfType(FloatActuator, UserBoolSensor, exclude=[sys.sens])
    assert sys.prog.actual_triggers == {sys.act, sys.a2}


def test_call_eval(mysys):
    call_eval = mysys.prog.active_condition.call_eval
    #assert call_eval(lambda x: x + 1, 1) == 2
    assert call_eval(1, 1) == 1

    assert call_eval(mysys.act, mysys.prog) == mysys.act._status
    assert call_eval(1, mysys.prog) == 1


class TestAbstractLogicCallable(object):

    #    @mock.patch.object(Attrib, '_give_triggers')
    #    @mock.patch.object(Empty, '_give_triggers')
    #    def test_collect(self, mock_empty, mock_attrib, mysys):
    #        mock_empty.return_value = [mysys.sens, 1]
    #        mock_attrib.return_value = [mysys.act, mysys.act, 2, 3]
    #        c = Empty()
    #        c2 = Attrib()
    #        assert c.collect_triggers() == {mysys.sens}
    #
    #        c._args = [c2]
    #        assert c.collect_triggers() == {mysys.sens, mysys.act}
    #
    #        mock_attrib.return_value = [mysys.act]
    #        assert c.collect_triggers() == {mysys.sens, mysys.act}
    #
    #        mock_attrib.return_value = [mysys.sens]
    #        assert c.collect_triggers() == {mysys.sens}
    #
    #        mock_attrib.return_value = [1, 2, 3, 4, 5]
    #        assert c.collect_triggers() == {mysys.sens}

    def test_collect_real(self, mysys):
        mysys.prog.on_activate = c = Run(SetStatus(mysys.act, mysys.sens), Changed(mysys.sens))
        assert c.triggers == {mysys.sens}
        assert c.targets == {mysys.act}

    def test_collect_real2(self, mysys):
        mysys.prog.on_activate = c = Anything(mysys.act, Anything(mysys.act, mysys.sens))
        assert c.triggers == {mysys.act, mysys.sens}
        c[1]._args.remove(mysys.sens)
        assert c.triggers == {mysys.act}
        c[1]._args.append(mysys.sens)
        assert c.triggers == {mysys.act, mysys.sens}

    @pytest.mark.parametrize('x,strver',
                             [
                                 (
                                     Empty(),
                                     'Empty()'
                                 ),
                                 (
                                     Attrib('hep', 'tstattr'),
                                     "Attrib('hep', 'tstattr')"
                                 ),
                                 (
                                     Eval('print time.{param}()', 'import time', param='time'),
                                     "Eval('print time.{param}()', 'import time', param='time')"
                                 ),
                                 (
                                     Method([1], 'pop'),
                                     "Method([1], 'pop')"
                                 ),
                                 (
                                     Method([1], 'pop', arg1='hep'),
                                     "Method([1], 'pop', arg1='hep')",
                                 ),
                                 (
                                     SetStatus(1, 2),
                                     "SetStatus(1, 2)",
                                 ), ])
    def test_give_str2(self, x, strver, sysloader):
        class syst(System):
            c1 = UserAnySensor(on_deactivate=SetStatus('c1', x))
        s = sysloader.new_system(syst)

        c2 = s.c1.on_deactivate[1]
        if sys.version_info < (3, 0):
            strver = re.sub(r"([= \(])(['\"])", lambda m: m.group(1) + 'u' + m.group(2), strver)
        assert c2.give_str() == strver
        assert repr(c2) == strver
        assert str(c2) == strver

    def test_cancel(self):
        c = Empty()
        c2 = mock.create_autospec(Empty)
        c2.triggers = {}
        c2.targets = {}
        c2.set_system = lambda x: None

        c._args = [c2, 1, 2]
        c.cancel(None)
        assert c2.cancel.call_count == 1
        c._args[1] = c2
        c.cancel(None)
        assert c2.cancel.call_count == 2


def test_empty():
    c = Empty()
    assert c.value is None
    assert c._args == []
    assert c.collect_targets() == set()
    assert c.collect_triggers() == set()
    assert c.objects == []
    assert c.call(prog) is None
    assert c.call(1) is None


def test_call(prog):

    l = [0, 1]
    prog.on_activate = c = Method(l, 'pop', 0, no_caller=True)
    assert c.call(prog) == 0
    assert c.call(prog) == 1
    with pytest.raises(IndexError):
        c.call(prog)

    class mycls(object):

        def fun(self, *args, **kwargs):
            return args, kwargs

    obj = mycls()
    prog.on_activate = c2 = Method(obj, 'fun', 0, no_caller=True)
    assert c2.call(prog) == ((0,), {})
    prog.on_activate = c3 = Method(obj, 'fun', 0, kwarg=1, no_caller=True)
    assert c3.call(prog) == ((0,), {'kwarg': 1})


def test_func(prog):
    flag = []

    def fnc(*args, **kwargs):
        assert args == (1,)
        assert kwargs == {'tst': 'hep'}
        flag.append(1)
    prog.on_deactivate = f = Func(fnc, 1, tst='hep')
    f.call(prog)
    assert flag
    prog.on_deactivate = f = Func(lambda: 1)
    f.call(prog)


def test_attrib(prog):
    l = [1, 2, 3]

    prog.on_activate = c = Attrib(l, "pop")
    assert c.call(prog)() == 3

    class mycls(object):
        a = 1
    obj = mycls()
    prog.on_activate = c2 = Attrib(obj, 'a')
    assert c2.call(prog) == 1

def test_attrib2(sysloader):
    class mysys2(System):
        class joku(object):
            joku = 2.0
        s = UserFloatSensor(default=3.0)
        a = FloatActuator()
        p = Program(active_condition=Attrib(joku, 'joku'))
        p2 = Program(active_condition=Attrib(p, 'active'))
        p3 = Program(active_condition=Value(True),
                     on_activate=SetStatus(a, Attrib(s, 'status', no_eval=True)))


    s = sysloader.new_system(mysys2)
    assert s.p.active
    assert s.p2.active
    s.flush()
    assert s.s in s.p3.actual_triggers
    assert s.a.program == s.p3
    assert s.a.status == 3.0


def test_onlytriggers(mysys):
    mysys.prog.on_activate = c = OnlyTriggers(mysys.act, mysys.sens)
    assert c.collect_targets() == set()
    assert c.collect_triggers() == {mysys.sens, mysys.act}


def test_log(caplog, mysys):
    mysys.prog.on_activate = c = Log("hephep")
    c.call(prog)
    assert "hephep" in caplog.text()  # logentries

    mysys.prog.on_activate = c = Log("hephep %s", 'hop')
    c.call(prog)
    assert 'hephep hop' in caplog.text()
    mysys.prog.on_activate = c = Log(mysys.sens)
    c.call(mysys.prog)
    assert str(mysys.sens._status) in caplog.text()

    caplog.setLevel(logging.INFO)
    mysys.prog.on_activate = c = Debug('debughep')
    c.call(prog)
    assert 'debughep' not in caplog.text()
    caplog.setLevel(logging.DEBUG)
    c.call(prog)
    assert 'debughep' in caplog.text()


@pytest.mark.parametrize('x,r', [
    (ToStr(1), '1'),
    (ToStr(1, 2, no_sub=True), '1 2'),
    (ToStr("hep"), 'hep'),
    (ToStr("hep", 1, no_sub=True), 'hep 1'),
    (ToStr("hep", "hop", no_sub=True), "hep hop"),
    (ToStr('Test {}', 1), 'Test 1'),
    (ToStr('Test {muu} {joku}', joku=1, muu=2), 'Test 2 1'),
])
def test_tostr(sysloader, x, r):
    class syst(System):
        c1 = UserAnySensor(on_activate=SetStatus('c1', x))
    s = sysloader.new_system(syst)
    assert s.c1.status == r


def test_eval():
    c = Eval('a+1', pre_exec='a=1')
    assert c.call(prog) == 2
    c = Eval('a+1+{b}', pre_exec='a=1', b=1)
    assert c.call(prog) == 3
    assert Eval('class hep:\n    pass').call(prog) == True


@pytest.mark.parametrize('x, r', [
    (Shell('true'), 0),
    (Shell('false'), 1),
    (Shell('somecommandthatcannotbefound'), 127),
    (Shell('echo hep', output=True), 'hep\n'),
    (Shell('cat', output=True, input='hep\n'), 'hep\n'),
])
def test_shell(x, r, sysloader):
    class mysys(System):
        p1 = UserAnySensor(on_activate=SetStatus('p1', x))
    s = sysloader.new_system(mysys)

    assert s.p1.status == r


@pytest.mark.parametrize('var', [1, 2, 0])
def test_setstatus(mysys, var):
    mysys.sens.set_status(1)  # set program active
    mysys.flush()
    assert mysys.prog.active
    s = SetStatus(mysys.act, var)
    s.setup_callable_system(mysys)
    s.call(mysys.prog)
    assert s.collect_targets() == {mysys.act}
    assert s.collect_triggers() == set()
    mysys.flush()
    assert mysys.act._status == var


@pytest.mark.parametrize('var', [0, 1, [1] * 2, [2, 3]])
def test_setvalues(mysys, var):
    assert not mysys.prog.active
    mysys.prog.on_activate = s = SetStatus([mysys.act, mysys.a2], var)
    assert s.collect_targets() == {mysys.act, mysys.a2}
    assert s.collect_triggers() == set()
    mysys.sens.set_status(1)  # set program active
    mysys.flush()
    assert mysys.prog.active
    var = [var] if not isinstance(var, list) else var
    assert mysys.act._status == var[0]
    assert mysys.a2._status == var[-1]


def test_setvalue_call(mysys):
    assert not mysys.prog.active

    def func():
        return [2, 3]
    mysys.prog.on_activate = s = SetStatus([mysys.act, mysys.a2], Func(func))
    assert s.collect_targets() == {mysys.act, mysys.a2}
    assert s.collect_triggers() == set()
    mysys.sens.set_status(1)  # set program active
    mysys.flush()
    assert mysys.prog.active
    assert mysys.act._status == 2
    assert mysys.a2._status == 3


def test_setvalue_call2(mysys):
    assert not mysys.prog.active

    def targets():
        return [mysys.act, mysys.a2]

    def values():
        return [2, mysys.sens]
    mysys.prog.on_activate = s = SetStatus(Func(targets), Func(values))
    mysys.prog.targets = {mysys.act, mysys.a2}
    mysys.sens.set_status(1)  # set program active
    mysys.flush()
    assert mysys.prog.active
    assert mysys.act._status == 2
    assert mysys.a2._status == mysys.sens.status


def test_func_arguments_translated(sysloader):
    returnvalue = []

    def testfunc(arg, test=1, test2=1):
        returnvalue.append((arg, test, test2))

    class ms(System):
        f = UserFloatSensor(active_condition=Value('f'), on_activate=Func(testfunc, 'f', test='f', test2=Mult(2, 'f')))

    s = sysloader.new_system(ms)
    s.f.status = 2.3
    s.flush()
    assert returnvalue[0] == (s.f.status, s.f.status, 2 * s.f.status)

# def test_waituntil_blocking(sysloader):
#    class ms(System):
#        s = UserFloatSensor()
#        f = UserFloatSensor(active_condition=Value('f'), on_activate=Run(SetStatus('t2', 1), Threaded(WaitUntil(s, SetStatus('t', 1), block=True))))
#        t = UserFloatSensor()
#        t2 = UserFloatSensor()
#    s = sysloader.new_system(ms)
#    s.f.status = 1
#    s.flush()
#    assert s.t2.status == 1
#    assert s.t.status == 0
#    s.s.status=1
#    s.flush()
#    assert s.t.status == 1


def test_waituntil_noblock(sysloader):
    class ms(System):
        s = UserFloatSensor()
        f = UserFloatSensor(active_condition=Value('f'), on_activate=Run(
            SetStatus('t2', 1), WaitUntil(s, SetStatus('t', 1))))
        t = UserFloatSensor()
        t2 = UserFloatSensor()
    s = sysloader.new_system(ms)
    s.f.status = 1
    s.flush()
    assert s.t2.status == 1
    assert s.t.status == 0
    s.s.status = 1
    s.flush()
    assert s.t.status == 1


def test_waituntil_two(sysloader):
    f1, f2 = [], []

    def func1():
        f1.append(1)

    def func2():
        f2.append(1)

    class ms(System):
        s = UserFloatSensor()
        s2 = UserFloatSensor()

        f = UserFloatSensor(active_condition=Value('f'),
                            on_activate=WaitUntil(s, Func(func1)))
        f2 = UserFloatSensor(active_condition=Value('f2'),
                             on_activate=WaitUntil(s2, Func(func2)))

    s = sysloader.new_system(ms)
    assert not f1
    assert not f2
    s.f.status = 1
    s.flush()
    assert not f1
    assert not f2
    s.s.status = 1
    s.flush()
    assert f1
    assert not f2
    s.f2.status = 1
    s.flush()
    assert f1
    assert not f2
    s.s2.status = 1
    s.flush()
    assert f1
    assert f2
    s.f.status = 0
    s.f2.status = 0
    s.s.status = 0
    s.s2.status = 0
    f1.pop(), f2.pop()
    s.flush()
    s.f.status = 1
    assert not f1 and not f2
    s.flush()
    s.f2.status = 1
    assert not f1 and not f2
    s.flush()
    s.s.status = 1
    s.flush()
    assert f1 and not f2
    s.s2.status = 1
    s.flush()
    assert f1 and f2


def test_waituntil_cancel(sysloader):
    class ms(System):
        s = UserFloatSensor()
        f = UserFloatSensor(active_condition=Value('f'), on_activate=Run(
            SetStatus('t2', 1), WaitUntil(s, SetStatus('t', 1))))
        t = UserFloatSensor()
        t2 = UserFloatSensor()
    s = sysloader.new_system(ms)
    s.f.status = 1
    s.flush()
    assert s.t2.status == 1
    assert s.t.status == 0
    s.f.status = 0  # deactivating program -> should cancel waituntil
    s.flush()
    assert s.t.status == 0
    s.s.status = 1
    s.flush()
    assert s.t.status == 0  # -> so that this is still unchanged
    s.f.status = 1
    s.flush()
    assert s.t.status == 1


def test_waituntil_musicserver_case(sysloader):
    """
        This test was created to debug/investigate my own program.
        I'll leave this as a test now that it works as expected.
    """
    called = []

    def start_playback(volume_status):
        called.append(volume_status)

    class ms(System):
        reset_mplayer = (
            If(Not('out_actual'),
               Run(
                SetStatus('launchtime', 1),
                WaitUntil('soundcard_ready',
                          SetStatus('volume', 0),
                          SetStatus('volume', 10),
                          SetStatus('launchtime', 0),
                          ),
            )
            ))
        start = UserBoolSensor(
            active_condition=Value('start'),
            on_activate=Run('reset_mplayer',
                            WaitUntil(And('soundcard_ready', Not('launchtime')),
                                      Func(start_playback, 'volume'),
                                      SetStatus('start', 0))))

        volume = UserIntSensor()

        out_actual = BoolActuator(
            active_condition=Value('out_actual'),
            on_activate=SetStatus('soundcard_ready', 1)
        )

        launchtime = UserBoolSensor()

        out_buf_prog = Program(
            on_update=SetStatus('out_buffer', 'launchtime')
        )

        out_buffer = BoolActuator(
            on_update=SetStatus(out_actual, 'out_buffer'),
        )

        soundcard_ready = UserBoolSensor()

    s = sysloader.new_system(ms)
    s.start.status = 1
    s.flush()
    assert s.start.status == 0
    assert 10 in called  # ensure that volume was set before starting playback
    assert s.soundcard_ready.status


def test_waituntil_func_called(sysloader):
    called = []

    def myfunc():
        called.append(1)

    class ms(System):
        s = UserFloatSensor(active_condition=Value('s'),
                            on_activate=WaitUntil(Not('f'), Func(myfunc))
                            )
        s2 = UserFloatSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=Run(
                SetStatus('t2', 1),
                WaitUntil(
                    And(s, s2),
                    SetStatus('t', 1),
                    SetStatus('f', 0)
                )
            )
        )
        t = UserFloatSensor()
        t2 = UserFloatSensor()
    s = sysloader.new_system(ms)
    s.f.status = 1
    s.flush()
    assert s.t2.status == 1
    assert s.t.status == 0
    s.s.status = 1
    s.flush()
    assert s.t.status == 0
    s.s2.status = 1
    s.flush()
    assert s.t.status == 1
    assert called


def test_waituntil_callable(sysloader):
    class ms(System):
        s = UserFloatSensor()
        s2 = UserFloatSensor()
        f = UserFloatSensor(active_condition=Value('f'), on_activate=Run(
            SetStatus('t2', 1), WaitUntil(And(s, s2), SetStatus('t', 1))))
        t = UserFloatSensor()
        t2 = UserFloatSensor()
    s = sysloader.new_system(ms)
    w = s.f.on_activate[1]
    s.f.status = 1
    s.flush()
    assert s.t2.status == 1
    assert s.t.status == 0
    s.s.status = 1
    s.flush()
    assert s.t.status == 0
    s.s2.status = 1
    assert w.get_state(s.f).callbacks
    s.flush()
    assert s.t.status == 1
    assert not w.get_state(s.f).callbacks


def test_while_actuator_condition(sysloader):
    called = []

    def myfunc():
        called.append(1)

    class ms(System):
        s = IntActuator(default=0)

        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(s < 3, Func(myfunc), SetStatus(s, Add(s, 1)), do_after=SetStatus('f', 0))
        )
    s = sysloader.new_system(ms)
    s.f.status = 1
    # Flushing system is not sufficient because While is threaded activity.
    time.sleep(0.5)
    assert s.f.status == 0
    #assert s.s.program_status(s.f) == 3
    assert s.s.status == 0
    assert len(called) == 3


def test_while_akvadimmer(sysloader):
    called = []

    def myfunc():
        called.append(1)

    class ms(System):
        warm_lamp_out = FloatActuator(default=0.)
        cold_lamp_out = FloatActuator(default=0.)

        fd_mpl = UserFloatSensor(description='multiplier', default=0.7)
        fd_thr = UserFloatSensor(description='threshold', default=0.1)
        fd_slp = UserFloatSensor(description='sleep time', default=0.0)

        dimmer = While(
            Or(
                Value('cold_lamp_out'),
                Value('warm_lamp_out'),
            ),
            SetStatus('cold_lamp_out', IfElse(Value('cold_lamp_out') > Value(fd_thr),
                                              Value('cold_lamp_out') * Value(fd_mpl), 0)),
            SetStatus('warm_lamp_out', IfElse(Value('warm_lamp_out') > Value(fd_thr),
                                              Value('warm_lamp_out') * Value(fd_mpl), 0)),
            Func(time.sleep, 'fd_slp'),
            Func(myfunc),
            do_after=Run(SetStatus(['akvadimmer', 'fade_out'], [0] * 2))
        )

        warm_preset_akva = UserFloatSensor(value_min=0., value_max=1., default=1, user_editable=False)
        cold_preset_akva = UserFloatSensor(value_min=0., value_max=1., default=1., user_editable=False)

        fade_out = UserBoolSensor(priority=3,
                                  active_condition=Value('fade_out'),
                                  on_activate=Run('dimmer')
                                  )

        akvadimmer = UserBoolSensor(priority=1.,  # lower priority light for aquarium (raspi1) remote use only
                                    active_condition=Value('akvadimmer'),
                                    on_activate=IfElse(
                                        # if lamp already on, do not activate this function at all
                                        Or('warm_lamp_out', 'cold_lamp_out'),
                                        SetStatus('akvadimmer', 0),
                                        Run(SetStatus('warm_lamp_out', 'warm_preset_akva'),
                                            SetStatus('cold_lamp_out', 'cold_preset_akva'),
                                            Run('dimmer'),
                                            )
                                    ))

    s = sysloader.new_system(ms)
    s.akvadimmer.status = 1
    # Flushing system is not sufficient because While is threaded activity.
    time.sleep(1.0)
    assert len(called) == 8
    assert s.dimmer[0].status == False
    assert s.akvadimmer.status == 0


def test_while_basic(sysloader):
    called = []

    def myfunc():
        called.append(1)

    class ms(System):
        s = UserIntSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(Less(s, 3), Func(myfunc), SetStatus(s, Add(s, 1)))
        )
    s = sysloader.new_system(ms)
    s.f.status = 1
    # Flushing system is not sufficient because While is threaded activity.
    time.sleep(0.5)
    assert s.s.status == 3
    assert len(called) == 3


def test_while_do_after(sysloader):
    called = []
    do_after_called = []

    def myfunc():
        called.append(1)

    def do_after():
        do_after_called.append(1)

    class ms(System):
        s = UserIntSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(Less(s, 3), Func(myfunc), SetStatus(s, Add(s, 1)), do_after=Func(do_after))
        )
    s = sysloader.new_system(ms)
    s.f.status = 1
    # Flushing system is not sufficient because While is threaded activity.
    time.sleep(0.5)
    assert s.s.status == 3
    assert len(called) == 3
    assert do_after_called


def test_while_cancel(sysloader, caplog):
    called = []

    def myfunc():
        time.sleep(0.1)
        called.append(1)

    class ms(System):
        s = UserIntSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(Value(True), Func(myfunc), SetStatus(s, Add(s, 1)))
        )
    s = sysloader.new_system(ms)
    w = s.f.on_activate
    assert not w.get_state(s.f).threads
    s.f.status = 1
    s.flush()
    assert w.get_state(s.f).threads
    s.f.status = 0  # : Deactivates program => cancels action
    s.flush()
    #assert w.get_state(s.f).threads[0]._cancel_while
    time.sleep(0.5)
    assert not w.get_state(s.f)
    assert 'Canceling While' in caplog.text()


def test_while_nested(sysloader):
    called = []

    def myfunc(s, s2):
        called.append((s, s2))

    class ms(System):
        s = UserIntSensor()
        s2 = UserIntSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(s < 3, While(s2 < 3, Func(myfunc, s, s2), SetStatus(s2, Add(s2, 1))), SetStatus(s, s + 1))
        )
    s = sysloader.new_system(ms)
    s.f.status = 1
    s.flush()
    time.sleep(0.5)
    assert s.s.status == 3
    assert s.s2.status == 3


def test_while_nested_cancel(sysloader):
    called = []

    def myfunc(s, s2):
        called.append((s, s2))
        time.sleep(0.1)

    class ms(System):
        s = UserIntSensor()
        s2 = UserIntSensor()
        f = UserFloatSensor(
            active_condition=Value('f'),
            on_activate=While(s < float('inf'),
                              While(s2 < float('inf'),
                                    Func(myfunc, s, s2),
                                    SetStatus(s2, Add(s2, 1))),
                              SetStatus(s, s + 1))
        )
    s = sysloader.new_system(ms)
    w1 = s.f.on_activate
    w2 = w1[1]
    s.f.status = 1
    s.flush()
    time.sleep(0.5)
    assert len(w1.get_state(s.f).threads) == 1
    #assert len(w2.get_state(s.f).threads)==1
    assert not w1.get_state(s.f).threads[0]._cancel_while
    #assert not w2.get_state(s.f).cancel
    s.f.status = 0
    s.flush()
    with w1._lock:
        assert (w1.get_state(s.f).threads and w1.get_state(s.f).threads[0]._cancel_while) or not w1.get_state(s.f)
    # There are many threads of second while, so this is not so easy to test
    # with w2._lock:
    #    assert (w2.get_state(s.f).cancel and w2.get_state(s.f).threads) or not w2.get_state(s.f)
    time.sleep(0.5)
    assert not w1.get_state(s.f)
    assert not w2.get_state(s.f)
    assert s.s.status > 0
    assert s.s2.status > 0


def get_musicserver():
    def func(arg1, arg2=None):
        print(arg1, arg2)  # getattr(arg2, 'name', None)

    class MusicServer_Tests(System):
        normal_volume = UserIntSensor(
            default=65,
            value_min=0,
            value_max=100,
        )
        piano_volume = UserIntSensor(
            default=30,
            value_min=0,
            value_max=100,
        )

        israspi = Value(True)
        email_sender = Empty()

        class UrlPlay(Group):
            reset_mplayer = Run(
                If(Not('out_actual'),
                   Func(func, 'launchtime<-1', TriggeredBy(), return_value=False),
                   SetStatus('launchtime', 1),
                   WaitUntil('soundcard_ready',
                             SetStatus('volume', 0),
                             SetStatus('volume', IfElse('piano_on', 'piano_volume', 'normal_volume')),
                             SetStatus('launchtime', 0),
                             Func(func, 'launchtime <-0'),
                             ),
                   ),
                Empty('mpc pause'),
                If('mplayer_pid',
                   Empty('kill mplayer...'),
                   SetStatus('mplayer_pid', 0)
                   )
            )

            mplayer_pid = UserIntSensor(user_editable=False)

            mplayer = UserStrSensor(
                active_condition=Value('mplayer'),
                exclude_triggers=['piano_volume', 'normal_volume'],  # launchtime', 'out_actual', 'soundcard_ready'],
                on_update=Run(
                    Func(func, 'mplayer Before triggered_by', TriggeredBy(), return_value=False),
                    If(TriggeredBy('mplayer'),
                       Run(reset_mplayer),
                       Func(func, 'mplayer waituntil init', TriggeredBy(), return_value=False),
                       WaitUntil(And('soundcard_ready', Not('launchtime')),
                                 Func(func, 'mplayer waituntil through'),
                                 SetStatus(mplayer_pid, 1)
                                 )
                       ),
                )
            )

        class Commands(Group):
            tags = 'web'

            radiodei = UserEventSensor(
                on_activate=SetStatus('mplayer', 'mms://mms.radiodei.fi/RadioDeiR2HF')
            )

            start = UserBoolSensor(
                active_condition=Value('start'),
                on_activate=Run('reset_mplayer',
                                WaitUntil(And('soundcard_ready', Not('launchtime')),
                                          Empty('mpc play'),
                                          SetStatus('start', 0))
                                )
            )

            stop = UserEventSensor(
                on_activate=Run('reset_mplayer', 'clear'),
            )

            volume = UserIntSensor(
                default=60,
                value_min=0,
                value_max=100,
            )

            manual_mode = UserBoolSensor(
                default=False,
            )

        class In(Group):
            soundcard_ready = UserBoolSensor(
                active_condition=Value('soundcard_ready'),
                on_activate=SetStatus('volume', IfElse('piano_on', 'piano_volume', 'normal_volume')),
            )

            playback_active = UserBoolSensor()

            set_piano_volume = Program(
                active_condition=And('piano_on', 'soundcard_ready'),
                on_activate=SetStatus('volume', 'piano_volume'),
                on_deactivate=SetStatus('volume', 'normal_volume')
            )

        class Out(Group):
            launchtime = UserBoolSensor()
            piano_on = UserBoolSensor()
            gmediarender_pid = UserIntSensor(default=0, user_editable=False)
            out_actual = BoolActuator(
                active_condition=Value('out_actual'),
                on_activate=SetStatus('gmediarender_pid', 1),
                on_deactivate=If('gmediarender_pid',
                                 Run(Empty(),
                                     SetStatus('gmediarender_pid', 0))),
            )

            out_buf_prog = Program(
                on_update=SetStatus('out_buffer', Or('launchtime', 'playback_active', 'piano_on', 'manual_mode'))
            )

            out_buffer = BoolActuator(
                safety_mode='both',
                change_delay=900.0,
                change_mode='falling',
                on_update=SetStatus(out_actual, 'out_buffer'),
            )
    return MusicServer_Tests


def test_musicserver_start(sysloader):
    s = sysloader.new_system(get_musicserver())
    assert not s.out_buffer.status
    s.start.status = True
    s.flush()
    assert s.out_buffer.status
    assert s.start.status
    assert s.launchtime.status
    s.soundcard_ready.status = 1
    # s.launchtime.status=False
    s.flush()
    assert not s.launchtime.status
    assert not s.start.status


def test_musicserver_radiodei(sysloader):
    s = sysloader.new_system(get_musicserver())
    reset_mplayer_waituntil_soundcard_ready = s.reset_mplayer[0][3]
    mplayer_on_update_waituntil = s.mplayer.on_update[1][3]

    assert s.mplayer.on_update.triggers == set()  # {s.piano_volume, s.normal_volume}
    assert mplayer_on_update_waituntil.triggers == set()  # s.launchtime, s.soundcard_ready}
    assert isinstance(mplayer_on_update_waituntil, WaitUntil)
    assert isinstance(reset_mplayer_waituntil_soundcard_ready, WaitUntil)
    assert not reset_mplayer_waituntil_soundcard_ready.state
    assert not mplayer_on_update_waituntil.state
    assert not s.out_buffer.status
    assert not s.out_actual.status
    assert not s.mplayer.status
    s.radiodei.status = True
    # mplayer <- url
    #    reset_mplayer
    #       launchtime <- 1
    #       waituntil soundcard_ready...
    #    waituntil soundcard_ready and not launchtime
    #
    s.flush()
    # there should be now a callback waiting in both of these waituntils.
    assert reset_mplayer_waituntil_soundcard_ready.state
    assert mplayer_on_update_waituntil.state
    assert s.mplayer.status  # contains url
    assert s.launchtime.status  # reset_mplayer switched this on
    assert s.out_buffer.status  # launchtime = 1 switched this on
    assert s.out_actual.status  # out_buffer turned this on
    # s.launchtime.status=False
    s.flush()
    s.soundcard_ready.status = 1
    # reset_mplayer should continue now and turn launchtime to 0
    s.flush()
    assert not s.launchtime.status
    assert s.out_buffer.status
    # also another waituntil should have run and set mplayer_pid <- 1
    assert s.mplayer_pid.status


def test_setattr(mysys):
    class mycls(object):
        a = 1
    a = mycls()
    mysys.prog.on_deactivate = c = SetAttr(a, b=2, c=3, d=mysys.sens)
    # c.setup_callable_system(mysys)
    c.call(prog)
    assert a.b == 2
    assert a.c == 3
    assert a.d == mysys.sens._status
    assert c.collect_triggers() == {mysys.sens}
    c._args[0] = mysys.act
    assert c.collect_targets() == {mysys.act}


def test_changed(mysys):
    mysys.prog.on_deactivate = c = Changed(mysys.sens)
    s = AbstractSensor(system=mysys, default=False)
    mysys.flush()
    assert s._status == False
    assert c.call(prog) == True
    assert c.call(prog) == False
    mysys.prog.on_update = Run(SetStatus(mysys.act, NEWVAL), SetStatus(s, c))
    mysys.sens.set_status(True)
    mysys.flush()
    assert s._status == True
    assert c.collect_targets() == set()
    assert c.collect_triggers() == {mysys.sens}


def test_swap(mysys):
    assert mysys.sens._status == 0
    assert mysys.act._status == 1.25
    mysys.prog.active_condition = Value(True)
    mysys.prog.update_condition = Value(True)
    mysys.prog.triggers = [mysys.sens]
    mysys.flush()
    assert mysys.sens._status == 0
    assert mysys.act._status == 2.0
    assert mysys.prog.active
    mysys.prog.on_update = Swap(mysys.act)
    mysys.flush()
    assert mysys.sens._status == 0
    assert mysys.act._status == 0
    mysys.sens.set_status(1)
    mysys.flush()
    assert mysys.act._status == 1
    mysys.sens.set_status(0)
    mysys.flush()
    assert mysys.act._status == 0
    #assert False


def test_run(prog):
    count = 0
    c1 = Exec('count += 1', namespace=locals())
    c = prog.on_deactivate = Run(c1, c1, c1)
    c.call(prog)
    assert c1._kwargs['namespace']['count'] == 3


def test_delay(caplog, mysys):
    c = Delay(0.05, Log('hep'))
    mysys.namespace['c'] = c
    c.call(prog)
    assert 'Scheduling' in caplog.text()
    time.sleep(1.5)
    print('2', caplog.text())
    assert 'Time is up' in caplog.text()


def test_delay_cancel(caplog, prog):
    c = Delay(1, Log('hep'))
    #mysys.namespace['c'] = c
    prog.on_activate = c
    c.call(prog)
    assert 'Scheduling' in caplog.text()
    assert c.get_state(prog).timers
    c.cancel(prog)
    time.sleep(2)
    # print caplog.text()
    assert not c.get_state(prog)
    assert 'Cancelling Delay' in caplog.text()
    assert "Time is up" not in caplog.text()


# def test_delay_cancel(caplog):
#    c = Delay(0.05, Log('hep'))
#    c.call(prog)
#    assert 'Scheduling' in caplog.text()
#    assert c._timer
#    c.cancel()
#    assert not c._timer
#    time.sleep(0.1)
#    assert 'Cancelling Delay' in caplog.text()
#    assert "Time is up" not in caplog.text()


def test_delay_secondtry(caplog, mysys):
    """
    Now the definition is to run always, not cancel
    """
    mysys.prog.on_deactivate = c = Delay(0.05, Log('hep'))
    c.call(prog)
    assert 'Scheduling' in caplog.text()
    c.call(prog)
    assert len(c.get_state(prog).timers) == 2
    time.sleep(1.5)
    assert "Time is up" in caplog.text()
    assert len(c.get_state(prog).timers) == 0


def test_ifelse(mysys):
    class mycls(object):
        r = 0
    itm1 = mycls()
    itm2 = mycls()

    def reset():
        itm1.r = 0
        itm2.r = 0
    mysys.prog.on_deactivate = c1 = SetAttr(itm1, r=1)
    mysys.prog.on_deactivate = c2 = SetAttr(itm2, r=1)

    mysys.namespace['i1'] = i = IfElse(0, c1)
    i.call(prog)
    assert itm1.r == 0

    mysys.namespace['i2'] = i = IfElse(1, c1)
    i.call(prog)
    assert itm1.r == 1
    reset()

    mysys.namespace['i3'] = i = IfElse(0, c1, c2)
    i.call(prog)
    assert itm1.r == 0
    assert itm2.r == 1
    reset()

    mysys.namespace['i4'] = i = IfElse(1, c1, c2)
    i.call(prog)
    assert itm1.r == 1
    assert itm2.r == 0

    mysys.namespace['i5'] = _if = IfElse(mysys.sens, mysys.act, mysys.a2)
    assert _if.collect_triggers() == {mysys.act, mysys.a2}
    assert _if.collect_targets() == {mysys.act, mysys.a2}


@pytest.mark.parametrize('var', list(range(3)))
def test_switch(var, mysys):
    i = mysys.namespace['i'] = Switch(var, 1, 2, 3)
    i.call(prog) == var + 1
    mysys.prog.on_deactivate = c = Switch(mysys.sens, mysys.act)
    assert c.collect_targets() == {mysys.act}
    assert c.collect_triggers() == {mysys.act}


def test_switch_dict(sysloader):
    class mysys(System):
        a = IntActuator()
        s = UserStrSensor(active_condition=Value('s'),
                          on_update=Switch('s', {'cmd1': SetStatus(a, 1),
                                                 'cmd2': SetStatus(a, 2)
                                                 }
                                           ),
                          )
    s = sysloader.new_system(mysys)
    assert not s.s.active
    s.s.status = 'cmd1'
    s.flush()
    assert s.s.active
    assert s.s.on_update.collect_targets() == {s.a}
    assert s.s.actual_targets == {s.a}
    assert s.a.program == s.s
    assert s.a.status == 1
    s.s.status = 'cmd2'
    s.flush()
    assert s.a.status == 2


def test_tryexcept(prog):
    e1 = Eval('aslkfdj')
    e2 = Eval('1')
    e3 = Eval('2')
    c = prog.on_deactivate = TryExcept(e1, e2)
    assert c.call(prog) == 1
    c = prog.on_deactivate = TryExcept(e2, e3)
    assert c.call(prog) == 1

nums = [1, 5, -2]


@pytest.mark.parametrize('x,r', [
    (Min(*nums), min(nums)),
    (Max(*nums), max(nums)),
    (Sum(*nums),  sum(nums)),
    (Product(*nums), 1 * 5 * -2),
    (Mult(*nums), 1 * 5 * -2),
    (Add(*nums), sum(nums)),
    (Div(1,5), 1/5),
    (Div(1.,5.), 1./5.),
])
def test_math(prog, x, r):
    prog.on_deactivate = c = x
    assert x.call(prog) == r


@pytest.mark.parametrize('x,r', [
    (Anything(0, 0, 0), True),
    (Anything(1, 0, 0), True),
    (Anything(1, 0, 1), True),
    (Or(1, 0, 1), True),
    (Or(0, 0, 0), False),
    (Or(1, 1, 1), True),

    (Or([0, 0, 0], 0, 1), True),
    (Or([0, 0, 1], 0, 0), True),

    (And(1, 1), True),
    (And(0, 1), False),
    (And(0, 0), False),
    (And([0, 0], 0), False),
    (And([1, 0], 0), False),
    (And([0, 0], 1), False),
    (And([1, 1], 0), False),
    (And([1, 1], 1), True),

    (Neg(0), 0),
    (Neg(1), -1),
    (Neg(-2), 2),
    (Not(0), True),
    (Not(1), False),
    (Equal(1, 1), True),
    (Equal(2, 1), False),
    (Equal(0, 0), True),

    (Less(0, 1), True),
    (Less(1, 1), False),
    (Less(1, 0), False),
    (More(0, 1), False),
    (More(1, 0), True),
    (More(1, 1), False),
    (Value(5), 5),
])
def test_logical(prog, x, r):
    prog.on_deactivate = x
    assert x.call(prog) == r


def test_logical2(prog):
    prog.on_deactivate = c = Neg(1, 1)
    with pytest.raises(RuntimeError):
        c.call(prog)
    c = prog.on_deactivate = Not(1, 1)
    with pytest.raises(RuntimeError):
        c.call(prog)


def test_logic_cmp():
    v1 = Value(1)
    v2 = Value(2)
    assert v1 * v2 == Product(v1, v2)
    assert v1 / v2 == Division(v1, v2)
    assert v1 + v2 == Sum(v1, v2)
    assert v1 - v2 == Sum(v1, Neg(v2))
    assert -v1 == Neg(v1)
    assert (v1 < v2) == Less(v1, v2)
    assert (v1 > v2) == More(v1, v2)


@pytest.fixture
def self_sys():
    class sys(System):
        s2 = UserBoolSensor()
        s = UserIntSensor(active_condition=Value('s'),
                          on_activate=SetStatus(['s', 's2'], [0, 1])
                          )
    s = sys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()


def test_self(self_sys):
    S = self_sys
    assert S.s.status == 0
    assert S.s2.status == False
    S.s.status = 1
    S.flush()
    assert S.s2.status == True
    assert S.s.status == 0


def test_self_delay(self_sys):
    S = self_sys
    S.s.on_activate = Delay(0.1, SetStatus('s', 0))
    assert S.s.status == 0
    S.s.status = 1
    S.flush()
    assert S.s.status == 1
    time.sleep(1.5)
    S.flush()
    assert S.s.status == 0


@pytest.mark.parametrize('x,r', [
    (RegexMatch(r'(\d*)(\w*)', '12test'), '12test'),
    (RegexSearch(r'(\d*)(\w*)', '12test'), '12'),

    (RegexMatch(r'(\d*)(\w*)', '12test', group=2), 'test'),
    (RegexSearch(r'(\d*)(\w*)', '12test', group=2), 'test'),


    (RegexMatch(r'testasfd', 'test'), ''),
    (RegexSearch(r'testasfd', 'test'), ''),

    (RegexMatch(r'heptest', 'heptest'), 'heptest'),
    (RegexSearch(r'heptest', 'heptest', group=0), 'heptest'),

    (RegexMatch(r'heptest1', 'heptest'), ''),
    (RegexSearch(r'heptest1', 'heptest'), ''),

    (RegexMatch(r'(hep)te(st1)', 'heptest1', group=1), 'hep'),
    (RegexSearch(r'(hep)te(st1)', 'heptest1', group=1), 'hep'),

    (RegexMatch(r'(hep)te(st1)', 'heptest1', group=2), 'st1'),
    (RegexSearch(r'(hep)te(st1)', 'heptest1', group=2), 'st1'),


])
def test_regex(x, r, sysloader):
    class mysys(System):
        p1 = UserAnySensor(on_activate=SetStatus('p1', x))
    s = sysloader.new_system(mysys)

    assert s.p1.status == r

# def test_regexmatch(sysloader):
#    class mysys(System):
#        p1 = UserBoolSensor(on_update=SetStatus('p1', RegexMatch('test', 'test')))
#        p2 = UserBoolSensor(on_update=SetStatus('p2', RegexMatch('testasfd', 'test')))
#        p3 = UserBoolSensor(on_update=SetStatus('p3', RegexMatch('^hep\w*$', 'heptest')))
#    s = sysloader.new_system(mysys)
#
#    assert s.p1.status == True
#    assert s.p2.status == False
#    assert s.p3.status == True


def test_triggeredby_basic(sysloader):
    class mysys(System):
        s1 = UserBoolSensor()
        s2 = UserBoolSensor()
        s3 = UserIntSensor()
        p = Program(active_condition=Or(s1, s2),
                    on_update=If(TriggeredBy(s1), SetStatus(s3, 1))
                    )

    s = sysloader.new_system(mysys)
    assert s.p.actual_triggers == {s.s1, s.s2}
    assert s.p.actual_targets == {s.s3}
    s.s2.status = 1
    s.flush()
    assert s.s3.status == 0
    s.s1.status = 1
    s.flush()
    assert s.s3.status == 1


def test_triggeredby_start(sysloader):
    l = []

    def fnc():
        l.append(1)

    class mysys(System):
        start = UserBoolSensor(
            active_condition=Value('start'),
            on_activate=SetStatus('mplayer', 'jotain')
        )
        mplayer = UserStrSensor(active_condition=Value('mplayer'),
                                on_update=If(TriggeredBy('mplayer'), Func(fnc)))
    s = sysloader.new_system(mysys)
    assert not l
    s.start.status = 1
    s.flush()
    assert l


def test_triggeredby_start2(sysloader):
    l = []

    def fnc():
        l.append(1)
        return 12

    class mysys(System):
        radiodei = UserEventSensor(
            on_activate=SetStatus('mplayer', 'mms://mms.radiodei.fi/RadioDeiR2HF')
        )

        reset_mplayer = Empty()

        mplayer_pid = UserIntSensor()
        mplayer = UserStrSensor(
            active_condition=Value('mplayer'),
            on_update=If(TriggeredBy('mplayer'),
                         Run(reset_mplayer),
                         SetStatus(mplayer_pid,
                                   Func(fnc))),
        )
    s = sysloader.new_system(mysys)
    assert not l
    s.radiodei.status = 1
    s.flush()
    assert l
    assert s.mplayer_pid.status == 12


def test_triggeredby_list(sysloader):
    class mysys(System):
        s1 = UserBoolSensor()
        s2 = UserBoolSensor()
        s3 = UserBoolSensor()
        s4 = UserBoolSensor()
        a = UserIntSensor()
        p = Program(active_condition=Or(s1, s2, s3, s4),
                    on_update=If(TriggeredBy(s1, s2), SetStatus(a, 1))
                    )

    s = sysloader.new_system(mysys)
    assert s.p.actual_triggers == {s.s1, s.s2, s.s3, s.s4}
    assert s.p.actual_targets == {s.a}
    s.s3.status = 1
    s.flush()
    assert s.a.status == 0
    s.s4.status = 1
    s.flush()
    assert s.a.status == 0
    s.s1.status = 1
    s.flush()
    assert s.a.status == 1


def test_triggeredby_triggername(sysloader):
    class mysys(System):
        s1 = UserBoolSensor()
        s2 = UserBoolSensor()
        s3 = UserBoolSensor()
        s4 = UserBoolSensor()
        a = UserIntSensor()
        p = Program(active_condition=Or(s1, s2, s3, s4),
                    on_update=If(Equal(TriggeredBy(), s1), SetStatus(a, 1))
                    )

    s = sysloader.new_system(mysys)
    assert s.p.actual_triggers == {s.s1, s.s2, s.s3, s.s4}
    assert s.p.actual_targets == {s.a}
    s.s3.status = 1
    s.flush()
    assert s.a.status == 0
    s.s4.status = 1
    s.flush()
    assert s.a.status == 0
    s.s1.status = 1
    s.flush()
    assert s.a.status == 1



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

"""
    Module for builtin Actuator classes
"""
from __future__ import division
from __future__ import unicode_literals

from threading import Thread
import time

from traits.api import Any, CBool, Instance, CInt, CFloat

from automate.common import threaded, get_modules_all
from automate.statusobject import AbstractActuator


class BoolActuator(AbstractActuator):

    """ Boolean valued actuator"""
    _status = CBool(transient=True)
    default = CBool


class IntActuator(AbstractActuator):

    """ Integer valued actuator"""
    _status = CInt(transient=True)
    default = CInt


class FloatActuator(AbstractActuator):

    """Floating point valued actuator"""
    _status = CFloat(transient=True)
    default = CFloat
    silent = CBool(True)


class AbstractInterpolatingActuator(FloatActuator):

    """
        Abstract base class for interpolating actuators.
    """
    #: How often to update status (as frequency)
    change_frequency = CFloat

    #: Slave actuator, that does the actual work (set .slave attribute to True in slave actuator)
    slave_actuator = Instance(AbstractActuator)

    _changethread = Any(transient=True)
    view = FloatActuator.view + ["change_frequency"]

    def _status_changed(self):
        if self._changethread and self._changethread.is_alive():
            return
        self._changethread = Thread(target=threaded(self.statuschanger),
                                    name="Changethread for " + self.name)
        self._changethread.start()


class ConstantSpeedActuator(AbstractInterpolatingActuator):

    """
        Change slave status with constant speed
    """

    #: Status change speed (change / second)
    speed = CFloat

    view = AbstractInterpolatingActuator.view + ["speed"]

    def statuschanger(self):
        if not self.slave_actuator:
            return
        while self.slave_actuator.status != self.status:
            delta = self.speed if self.status > self.slave_actuator.status else -self.speed
            newstatus = self.slave_actuator.status + delta / self.change_frequency
            if (delta > 0) == (newstatus > self.status) or abs(newstatus - self.status) < delta / self.change_frequency:
                self.slave_actuator.status = self.status
            else:
                self.slave_actuator.status = newstatus
            time.sleep(1. / self.change_frequency)


class ConstantTimeActuator(ConstantSpeedActuator):

    """
    Change slave status in constant time
    """

    #: Time that is needed for change
    change_time = CFloat

    view = ConstantSpeedActuator.view + ["change_time"]

    def _status_changed(self):
        self.speed = abs(self.slave_actuator.status - self.status) / self.change_time
        if self._changethread and self._changethread.is_alive():
            return
        self._changethread = Thread(target=threaded(self.statuschanger),
                                    name="Changethread for " + self.name)
        self._changethread.start()


__all__ = get_modules_all(AbstractActuator, locals())
